# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Helpers for Python 2/3 compatibility.
"""

from json import dumps

from twisted.python.compat import unicode

def dumps_bytes(obj):
    """
    Serialize ``obj`` to JSON formatted ``bytes``.
    """
    b = dumps(obj)
    if isinstance(b, unicode):
        b = b.encode("ascii")
    return b



def native_string_to_bytes(s, encoding="ascii", errors="strict"):
    """
    Ensure that the native string ``s`` is converted to ``bytes``.
    """
    if not isinstance(s, str):
        raise TypeError("{} must be type str, not {}".format(s, type(s)))
    if str is bytes:
        # Python 2
        return s
    else:
        # Python 3
        return s.encode(encoding=encoding, errors=errors)



def native_string_to_unicode(s, encoding="ascii", errors="strict"):
    """
    Ensure that the native string ``s`` is converted to ``unicode``.
    """
    if not isinstance(s, str):
        raise TypeError("{} must be type str, not {}".format(s, type(s)))
    if str is unicode:
        # Python 3
        return s
    else:
        # Python 2
        return s.decode(encoding=encoding, errors=errors)



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Functional structures for representing different kinds of Kubernetes
state.
"""

from uuid import uuid4
from json import loads

import attr

from zope.interface import implementer

from pyrsistent import mutant

from twisted.python.compat import iteritems, unicode
from twisted.python.filepath import FilePath

from . import UnrecognizedVersion, UnrecognizedKind, IObject
from ._swagger import NoSuchDefinition, Swagger, VersionedPClasses

@attr.s(frozen=True)
class _KubernetesDataModel(object):
    """
    A representation of txkube's understanding of the data model of some
    particular version of Kubernetes.
    """
    spec = attr.ib()
    version_type = attr.ib()
    version = attr.ib()

    v1 = attr.ib()
    v1beta1 = attr.ib()

    @classmethod
    def from_path(cls, path, version_type_name, version_details, v1, v1beta1):
        return cls.from_swagger(
            Swagger.from_path(path),
            version_type_name,
            version_details,
            v1,
            v1beta1,
        )


    @classmethod
    def from_swagger(cls, swagger, version_type_name, version_details, v1, v1beta1):
        spec = VersionedPClasses.transform_definitions(swagger)
        v1 = VersionedPClasses(spec, v1)
        v1beta1 = VersionedPClasses(spec, v1beta1)
        version_type = spec.pclass_for_definition(version_type_name)
        version = version_type(**version_details)
        model = cls(
            spec=spec,
            version_type=version_type,
            version=version,
            v1=v1,
            v1beta1=v1beta1,
        )
        define_behaviors(model)
        return model


    @mutant
    def iobject_from_raw(self, obj):
        """
        Load an object of unspecified type from the raw representation of it.

        :raise KeyError: If the kind of object is unsupported.

        :return IObject: The loaded object.
        """
        versions = {
            version: v
            for v in (self.v1, self.v1beta1)
            for version in v.versions
        }
        versions.update({
            "v1": self.v1,
            "v1beta1": self.v1beta1,
        })
        kind = obj[u"kind"]
        apiVersion = _unmutilate(obj[u"apiVersion"])
        try:
            v = versions[apiVersion]
        except KeyError:
            raise UnrecognizedVersion(apiVersion, obj)
        try:
            cls = getattr(v, kind)
        except AttributeError:
            raise UnrecognizedKind(apiVersion, kind, obj)
        others = obj.discard(u"kind").discard(u"apiVersion")
        return cls.create(others)


    def iobject_to_raw(self, obj):
        result = obj.serialize()
        result.update({
            u"kind": obj.kind,
            u"apiVersion": _mutilate(obj.apiVersion),
        })
        return result



def openapi_to_data_model(openapi):
    version = openapi[u"info"][u"version"]
    if version == u"unversioned":
        # Kubernetes 1.5.x probably.
        return _openapi_to_v1_5_data_model(openapi)
    elif version.startswith(u"v1.6."):
        return _openapi_to_v1_6_data_model(openapi)
    elif version.startswith(u"v1.7."):
        return _openapi_to_v1_7_data_model(openapi)
    else:
        # Optimistic...
        return _openapi_to_newest_data_model(openapi)


def _openapi_to_v1_5_data_model(openapi):
    base = loads(
        FilePath(__file__).sibling("extra-1.5.json").getContent()
    )
    for k, v in iteritems(openapi):
        if isinstance(v, dict):
            base.setdefault(k, {}).update(v)
        else:
            base[k] = v

    return _KubernetesDataModel.from_swagger(
        Swagger.from_document(base),
        u"version.Info", dict(
            major=u"1",
            minor=u"5",
            gitVersion=u"",
            gitCommit=u"",
            gitTreeState=u"",
            buildDate=u"",
            goVersion=u"",
            compiler=u"",
            platform=u"",
        ),
        {u"v1"},
        {u"v1beta1"},
    )

def _openapi_to_v1_6_data_model(openapi):
    return _KubernetesDataModel.from_swagger(
        Swagger.from_document(openapi),
        u"io.k8s.apimachinery.pkg.version.Info", dict(
            major=u"1",
            minor=u"6",
            gitVersion=u"",
            gitCommit=u"",
            gitTreeState=u"",
            buildDate=u"",
            goVersion=u"",
            compiler=u"",
            platform=u"",
        ),
        {
            u"io.k8s.kubernetes.pkg.api.v1",
            u"io.k8s.apimachinery.pkg.apis.meta.v1",
        },
        {
            u"io.k8s.kubernetes.pkg.apis.extensions.v1beta1",
            u"io.k8s.kubernetes.pkg.apis.certificates.v1beta1",
        },
    )

# 1.6 and 1.7 happen to look similar enough that no *tested* functionality
# requires us to differentiate between them here.
_openapi_to_v1_7_data_model = _openapi_to_v1_6_data_model

# Keep this up to date with whatever the newest thing we know about is...
_openapi_to_newest_data_model =  _openapi_to_v1_7_data_model


def set_if_none(desired_value):
    """
    Create a transformer which sets the given value if it finds ``None`` as
    the current value, otherwise leaves the current value alone.
    """
    def transform(current_value):
        if current_value is None:
            return desired_value
        return current_value
    return transform



def object_sort_key(obj):
    """
    Define a predictable sort ordering for Kubernetes objects.

    This should be the same ordering that Kubernetes itself imposes.
    """
    return (
        # Not all objects have a namespace.
        getattr(obj.metadata, "namespace", None),
        obj.metadata.name,
    )



def required_unique(objects, key):
    """
    A pyrsistent invariant which requires all objects in the given iterable to
    have a unique key.

    :param objects: The objects to check.
    :param key: A one-argument callable to compute the key of an object.

    :return: An invariant failure if any two or more objects have the same key
        computed.  An invariant success otherwise.
    """
    keys = {}
    duplicate = set()
    for k in map(key, objects):
        keys[k] = keys.get(k, 0) + 1
        if keys[k] > 1:
            duplicate.add(k)
    if duplicate:
        return (False, u"Duplicate object keys: {}".format(duplicate))
    return (True, u"")



def add(value):
    def evolver(pvector):
        return sorted(pvector.append(value), key=object_sort_key)
    return evolver



def remove(value):
    def evolver(pset):
        return pset.remove(value)
    return evolver



class _List(object):
    def __invariant__(self):
        return required_unique(self.items, object_sort_key)


    @staticmethod
    def __new__(cls, **kwargs):
        # The Kubernetes Swagger schema for Lists claims items is a *required*
        # *array*.  However, it is frequently None/null instead.  Hack around
        # such values, turning them into the empty sequence.
        #
        # It might be better to fix this with a schema overlay instead - eg,
        # with a tweak to mark them as optional instead of required.
        items = kwargs.get(u"items", None)
        if items is None:
            kwargs[u"items"] = ()
        return super(_List, cls).__new__(cls, **kwargs)


    def item_by_name(self, name):
        """
        Find an item in this collection by its name metadata.

        :param unicode name: The name of the object for which to search.

        :raise KeyError: If no object matching the given name is found.
        :return IObject: The object with the matching name.
        """
        for obj in self.items:
            if obj.metadata.name == name:
                return obj
        raise KeyError(name)


    def add(self, obj):
        return self.transform([u"items"], add(obj))


    def remove(self, obj):
        return self.transform([u"items"], remove(obj))


    def replace(self, old, new):
        return self.transform(
            [u"items"], remove(old),
            [u"items"], add(new),
        )



def define_behaviors(model):
    def add_allowing_missing(v):
        def adder(cls):
            try:
                return v.add_behavior_for_pclass(cls)
            except NoSuchDefinition:
                return None
        return adder

    @add_allowing_missing(model.v1)
    class NamespaceStatus(object):
        """
        ``NamespaceStatus`` instances model `Kubernetes namespace status
        <https://kubernetes.io/docs/api-reference/v1/definitions/#_v1_namespacestatus>`_.
        """
        @classmethod
        def active(cls):
            return cls(phase=u"Active")


        @classmethod
        def terminating(cls):
            return cls(phase=u"Terminating")



    @add_allowing_missing(model.v1)
    @implementer(IObject)
    class Namespace(object):
        """
        ``Namespace`` instances model `Kubernetes namespaces
        <https://kubernetes.io/docs/user-guide/namespaces/>`_.
        """
        @classmethod
        def default(cls):
            """
            Get the default namespace.
            """
            return cls(metadata=model.v1.ObjectMeta(name=u"default"))


        def fill_defaults(self):
            return self.transform(
                # TODO Also creationTimestamp, resourceVersion, maybe selfLink.
                # Also, should this clobber existing values or leave them alone?
                # See https://github.com/LeastAuthority/txkube/issues/36
                [u"metadata", u"uid"], unicode(uuid4()),
                [u"status"], model.v1.NamespaceStatus.active(),
            )


        def delete_from(self, collection):
            # TODO: deletionTimestamp?  Terminating status?  See #24
            return collection.remove(self)



    @add_allowing_missing(model.v1)
    @implementer(IObject)
    class NamespaceList(_List):
        pass



    @add_allowing_missing(model.v1)
    @implementer(IObject)
    class ConfigMap(object):
        """
        ``ConfigMap`` instances model `ConfigMap objects
        <https://kubernetes.io/docs/api-reference/v1/definitions/#_v1_configmap>`_.
        """
        def fill_defaults(self):
            # TODO Surely some stuff to fill.
            # See https://github.com/LeastAuthority/txkube/issues/36
            return self


        def delete_from(self, collection):
            return collection.remove(self)



    @add_allowing_missing(model.v1)
    @implementer(IObject)
    class ConfigMapList(_List):
            pass



    @add_allowing_missing(model.v1)
    @implementer(IObject)
    class Service(object):
        """
        ``Service`` instances model `Service objects
        <https://kubernetes.io/docs/api-reference/v1/definitions/#_v1_service>`_.
        """
        def fill_defaults(self):
            # TODO Surely some stuff to fill.
            # See https://github.com/LeastAuthority/txkube/issues/36
            return self


        def delete_from(self, collection):
            return collection.remove(self)



    @add_allowing_missing(model.v1)
    @implementer(IObject)
    class ServiceList(_List):
        pass



    @add_allowing_missing(model.v1beta1)
    @implementer(IObject)
    class Deployment(object):
        """
        ``Deployment`` instances model `Deployment objects
        <https://kubernetes.io/docs/api-reference/extensions/v1beta1/definitions/#_v1beta1_deployment>`_.
        """
        def fill_defaults(self):
            # Copying apparent Kubernetes behavior.
            return self.transform(
                [u"metadata", u"labels"],
                set_if_none(self.spec.template.metadata.labels),
            )


        def delete_from(self, collection):
            return collection.remove(self)



    @add_allowing_missing(model.v1beta1)
    @implementer(IObject)
    class DeploymentList(_List):
        pass



    @add_allowing_missing(model.v1beta1)
    @implementer(IObject)
    class ReplicaSet(object):
        """
        ``ReplicaSet`` instances model `ReplicaSet objects
        <https://kubernetes.io/docs/concepts/workloads/controllers/replicaset/>`_.
        """
        def fill_defaults(self):
            return self


        def delete_from(self, collection):
            return collection.remove(self)



    @add_allowing_missing(model.v1beta1)
    @implementer(IObject)
    class ReplicaSetList(_List):
        pass



    @add_allowing_missing(model.v1)
    @implementer(IObject)
    class Pod(object):
        """
        ``Pod`` instances model `Pod objects
        <https://kubernetes.io/docs/api-reference/v1/definitions/#_v1_pod>`_.
        """
        def fill_defaults(self):
            return self


        def delete_from(self, collection):
            return collection.remove(self)



    @add_allowing_missing(model.v1)
    @implementer(IObject)
    class PodList(_List):
        pass



def _mutilate(version):
    try:
        group, version = version.rsplit(u".", 1)
    except ValueError:
        pass
    if version == u"v1beta1":
        return u"extensions/v1beta1"
    return version



def _unmutilate(version):
    if version.startswith(u"extensions/"):
        return version[len(u"extensions/"):]
    return version


with FilePath(__file__).sibling(u"kubernetes-1.5.json").open() as v1_5_file:
    v1_5_model = openapi_to_data_model(loads(v1_5_file.read()))

iobject_to_raw = v1_5_model.iobject_to_raw
iobject_from_raw = v1_5_model.iobject_from_raw
v1 = v1_5_model.v1
v1beta1 = v1_5_model.v1beta1



# Copyright Least Authority Enterprises.
# See LICENSE for details.

import os
from itertools import count, islice
from uuid import uuid4

from pykube import KubeConfig

import pem

import attr

from pyrsistent import InvariantException

from hypothesis import given

from fixtures import TempDir

from zope.interface.verify import verifyObject

from testtools import ExpectedException
from testtools.matchers import (
    AfterPreprocessing, Equals, Contains, IsInstance, raises
)
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.backends import default_backend

from zope.interface import implementer

from twisted.python.compat import unicode
from twisted.python.filepath import FilePath
from twisted.internet.address import IPv4Address
from twisted.internet.error import DNSLookupError
from twisted.internet.interfaces import (
    IHostResolution,
    IReactorPluggableNameResolver,
    IOpenSSLClientConnectionCreator,
)
from twisted.internet.protocol import Factory
from twisted.web.iweb import IPolicyForHTTPS
from twisted.web.http_headers import Headers
from twisted.test.iosim import ConnectionCompleter
from twisted.test.proto_helpers import AccumulatingProtocol, MemoryReactorClock

from ..testing import TestCase, assertNoResult, cert
from ..testing.strategies import (
    dns_subdomains,
    port_numbers,
)

from .._authentication import (
    ClientCertificatePolicyForHTTPS,
    NetLocation,
    Certificates,
    Chain,
    pairwise,
    https_policy_from_config,
)
from .. import authenticate_with_serviceaccount
from ._compat import encode_environ

# Just an arbitrary certificate pulled off the internet.  Details ought not
# matter.  Retrieved using:
#
#    $ openssl s_client -showcerts -connect google.com:443
#
_CA_CERT_PEM = b"""\
-----BEGIN CERTIFICATE-----
MIIDfTCCAuagAwIBAgIDErvmMA0GCSqGSIb3DQEBBQUAME4xCzAJBgNVBAYTAlVT
MRAwDgYDVQQKEwdFcXVpZmF4MS0wKwYDVQQLEyRFcXVpZmF4IFNlY3VyZSBDZXJ0
aWZpY2F0ZSBBdXRob3JpdHkwHhcNMDIwNTIxMDQwMDAwWhcNMTgwODIxMDQwMDAw
WjBCMQswCQYDVQQGEwJVUzEWMBQGA1UEChMNR2VvVHJ1c3QgSW5jLjEbMBkGA1UE
AxMSR2VvVHJ1c3QgR2xvYmFsIENBMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIB
CgKCAQEA2swYYzD99BcjGlZ+W988bDjkcbd4kdS8odhM+KhDtgPpTSEHCIjaWC9m
OSm9BXiLnTjoBbdqfnGk5sRgprDvgOSJKA+eJdbtg/OtppHHmMlCGDUUna2YRpIu
T8rxh0PBFpVXLVDviS2Aelet8u5fa9IAjbkU+BQVNdnARqN7csiRv8lVK83Qlz6c
JmTM386DGXHKTubU1XupGc1V3sjs0l44U+VcT4wt/lAjNvxm5suOpDkZALeVAjmR
Cw7+OC7RHQWa9k0+bw8HHa8sHo9gOeL6NlMTOdReJivbPagUvTLrGAMoUgRx5asz
PeE4uwc2hGKceeoWMPRfwCvocWvk+QIDAQABo4HwMIHtMB8GA1UdIwQYMBaAFEjm
aPkr0rKV10fYIyAQTzOYkJ/UMB0GA1UdDgQWBBTAephojYn7qwVkDBF9qn1luMrM
TjAPBgNVHRMBAf8EBTADAQH/MA4GA1UdDwEB/wQEAwIBBjA6BgNVHR8EMzAxMC+g
LaArhilodHRwOi8vY3JsLmdlb3RydXN0LmNvbS9jcmxzL3NlY3VyZWNhLmNybDBO
BgNVHSAERzBFMEMGBFUdIAAwOzA5BggrBgEFBQcCARYtaHR0cHM6Ly93d3cuZ2Vv
dHJ1c3QuY29tL3Jlc291cmNlcy9yZXBvc2l0b3J5MA0GCSqGSIb3DQEBBQUAA4GB
AHbhEm5OSxYShjAGsoEIz/AIx8dxfmbuwu3UOx//8PDITtZDOLC5MH0Y0FWDomrL
NhGc6Ehmo21/uBPUR/6LWlxz/K7ZGzIZOKuXNBSqltLroxwUCEm2u+WR74M26x1W
b8ravHNjkOR/ez4iyz0H7V84dJzjA1BOoa+Y7mHyhD8S
-----END CERTIFICATE-----
"""

# Let hostname u"example.invalid" map to an
# IPv4 address in the TEST-NET range.
HOST_MAP = {
    u"example.invalid.": "192.0.2.2"
}

def create_reactor():
    """
    Twisted 17.1.0 and higher requires a reactor which implements
    ``IReactorPluggableNameResolver``.
    """

    @implementer(IHostResolution)
    @attr.s
    class Resolution(object):
        name = attr.ib()

    class _FakeResolver(object):

        def resolveHostName(self, resolutionReceiver, hostName, *args,  **kwargs):
            portNumber = kwargs.pop('portNumber')
            r = Resolution(name=hostName)

            resolutionReceiver.resolutionBegan(r)
            if hostName in HOST_MAP:
                resolutionReceiver.addressResolved(
                    IPv4Address('TCP', HOST_MAP[hostName], portNumber))
            resolutionReceiver.resolutionComplete()
            return r

    @implementer(IReactorPluggableNameResolver)
    class _ResolvingMemoryClockReactor(MemoryReactorClock):
        nameResolver = _FakeResolver()

    return _ResolvingMemoryClockReactor()



class AuthenticateWithServiceAccountTests(TestCase):
    """
    Tests for ``authenticate_with_serviceaccount``.
    """
    def _authorized_request(self, token, headers,
                            kubernetes_host=b"example.invalid."):
        """
        Get an agent using ``authenticate_with_serviceaccount`` and issue a
        request with it.

        :return bytes: The bytes of the request the agent issues.
        """
        server = AccumulatingProtocol()
        factory = Factory.forProtocol(lambda: server)
        factory.protocolConnectionMade = None

        reactor = create_reactor()
        reactor.listenTCP(80, factory)

        t = FilePath(self.useFixture(TempDir()).path)
        t = t.asBytesMode()
        serviceaccount = t.child(b"serviceaccount")
        serviceaccount.makedirs()

        serviceaccount.child(b"ca.crt").setContent(_CA_CERT_PEM)
        serviceaccount.child(b"token").setContent(token)

        environ = encode_environ(
            {
                u"KUBERNETES_SERVICE_HOST": kubernetes_host.decode("ascii"),
                u"KUBERNETES_SERVICE_PORT": u"443"
            })
        self.patch(os, "environ", environ)

        agent = authenticate_with_serviceaccount(
            reactor, path=serviceaccount.asTextMode().path,
        )

        d = agent.request(b"GET", b"http://" + kubernetes_host, headers)
        assertNoResult(self, d)
        [(host, port, factory, _, _)] = reactor.tcpClients

        addr = HOST_MAP.get(kubernetes_host.decode("ascii"), None)
        self.expectThat((host, port), Equals((addr, 80)))

        pump = ConnectionCompleter(reactor).succeedOnce()
        pump.pump()

        return server.data


    def test_bearer_token_authorization(self):
        """
        The ``IAgent`` returned adds an *Authorization* header to each request it
        issues.  The header includes the bearer token from the service account
        file.
        """
        token = str(uuid4())
        if isinstance(token, unicode):
            token = token.encode("ascii")
        request_bytes = self._authorized_request(token=token, headers=None)

        # Sure would be nice to have an HTTP parser.
        self.assertThat(
            request_bytes,
            Contains(b"Authorization: Bearer " + token),
        )


    def test_hostname_does_not_resolve(self):
        """
        Specifying a hostname which cannot be resolved to an
        IP address will result in an ``DNSLookupError``.
        """
        with ExpectedException(DNSLookupError, "DNS lookup failed: no results "
                               "for hostname lookup: doesnotresolve."):
            self._authorized_request(
                token=b"test",
                headers=Headers({}),
                kubernetes_host=b"doesnotresolve"
            )


    def test_other_headers_preserved(self):
        """
        Other headers passed to the ``IAgent.request`` implementation are also
        sent in the request.
        """
        token = str(uuid4())
        if isinstance(token, unicode):
            token = token.encode("ascii")
        headers = Headers({u"foo": [u"bar"]})
        request_bytes = self._authorized_request(token=token, headers=headers)
        self.expectThat(
            request_bytes,
            Contains(b"Authorization: Bearer " + token),
        )
        self.expectThat(
            request_bytes,
            Contains(b"Foo: bar"),
        )



class HTTPSPolicyFromConfigTests(TestCase):
    """
    Tests for ``https_policy_from_config``.
    """
    def test_policy(self):
        """
        ``https_policy_from_config`` returns a ``ClientCertificatePolicyForHTTPS``
        with no credentials but with trust roots taken from the Kubernetes
        *serviceaccount* directory it is pointed at.  It also respects
        *KUBERNETES_...* environment variables to identify the address of the
        server.
        """
        t = FilePath(self.useFixture(TempDir()).path)
        t = t.asBytesMode()
        serviceaccount = t.child(b"serviceaccount")
        serviceaccount.makedirs()

        serviceaccount.child(b"ca.crt").setContent(_CA_CERT_PEM)
        serviceaccount.child(b"token").setContent(b"token")

        netloc = NetLocation(host=u"example.invalid", port=443)
        environ = encode_environ({
                u"KUBERNETES_SERVICE_HOST": netloc.host,
                u"KUBERNETES_SERVICE_PORT": u"{}".format(netloc.port),
        })
        self.patch(os, "environ", environ)

        config = KubeConfig.from_service_account(path=serviceaccount.asTextMode().path)

        policy = https_policy_from_config(config)
        self.expectThat(
            policy,
            Equals(
                ClientCertificatePolicyForHTTPS(
                    credentials={},
                    trust_roots={
                        netloc: pem.parse(_CA_CERT_PEM)[0],
                    },
                ),
            ),
        )


    def test_missing_ca_certificate(self):
        """
        If no CA certificate is found in the service account directory,
        ``https_policy_from_config`` raises ``ValueError``.
        """
        t = FilePath(self.useFixture(TempDir()).path)
        t = t.asBytesMode()
        serviceaccount = t.child(b"serviceaccount")
        serviceaccount.makedirs()

        serviceaccount.child(b"ca.crt").setContent(b"not a cert pem")
        serviceaccount.child(b"token").setContent(b"token")

        environ = encode_environ({
            u"KUBERNETES_SERVICE_HOST": u"example.invalid.",
            u"KUBERNETES_SERVICE_PORT": u"443",
        })
        self.patch(os, "environ", environ)

        config = KubeConfig.from_service_account(path=serviceaccount.asTextMode().path)
        self.assertThat(
            lambda: https_policy_from_config(config),
            raises(ValueError("No certificate authority certificate found.")),
        )


    def test_bad_ca_certificate(self):
        """
        If no CA certificate is found in the service account directory,
        ``https_policy_from_config`` raises ``ValueError``.
        """
        t = FilePath(self.useFixture(TempDir()).path)
        t = t.asBytesMode()
        serviceaccount = t.child(b"serviceaccount")
        serviceaccount.makedirs()

        serviceaccount.child(b"ca.crt").setContent(
            b"-----BEGIN CERTIFICATE-----\n"
            b"not a cert pem\n"
            b"-----END CERTIFICATE-----\n"
        )
        serviceaccount.child(b"token").setContent(b"token")

        environ = encode_environ({
            u"KUBERNETES_SERVICE_HOST": u"example.invalid.",
            u"KUBERNETES_SERVICE_PORT": u"443",
        })
        self.patch(os, "environ", environ)

        config = KubeConfig.from_service_account(path=serviceaccount.asTextMode().path)
        self.assertThat(
            lambda: https_policy_from_config(config),
            raises(ValueError(
                "Invalid certificate authority certificate found.",
                "[('PEM routines', 'PEM_read_bio', 'bad base64 decode')]",
            )),
        )



class ClientCertificatePolicyForHTTPSTests(TestCase):
    """
    Tests for ``ClientCertificatePolicyForHTTPS``.
    """
    def test_interface(self):
        """
        ``ClientCertificatePolicyForHTTPS`` instances provide ``IPolicyForHTTPS``.
        """
        policy = ClientCertificatePolicyForHTTPS(
            credentials={},
            trust_roots={},
        )
        verifyObject(IPolicyForHTTPS, policy)


    @given(dns_subdomains(), dns_subdomains(), port_numbers(), port_numbers())
    def test_creatorForNetLoc_interface(self, host_known, host_used, port_known, port_used):
        """
        ``ClientCertificatePolicyForHTTPS.creatorForNetloc`` returns an object
        that provides ``IOpenSSLClientConnectionCreator``.
        """
        netloc = NetLocation(host=host_known, port=port_known)
        cert = pem.parse(_CA_CERT_PEM)[0]

        policy = ClientCertificatePolicyForHTTPS(
            credentials={},
            trust_roots={
                netloc: cert,
            },
        )
        creator = policy.creatorForNetloc(
            host_used.encode("ascii"),
            port_used,
        )
        verifyObject(IOpenSSLClientConnectionCreator, creator)



class PairwiseTests(TestCase):
    """
    Tests for ``pairwise``.
    """
    def test_pairs(self):
        a = object()
        b = object()
        c = object()
        d = object()

        self.expectThat(
            pairwise([]),
            AfterPreprocessing(list, Equals([])),
        )
        self.expectThat(
            pairwise([a]),
            AfterPreprocessing(list, Equals([])),
        )
        self.expectThat(
            pairwise([a, b]),
            AfterPreprocessing(list, Equals([(a, b)])),
        )

        self.expectThat(
            pairwise([a, b, c]),
            AfterPreprocessing(list, Equals([(a, b), (b, c)])),
        )
        self.expectThat(
            pairwise([a, b, c, d]),
            AfterPreprocessing(list, Equals([(a, b), (b, c), (c, d)])),
        )


    def test_lazy(self):
        """
        ``pairwise`` only consumes as much of its iterable argument as necessary
        to satisfy iteration of its own result.
        """
        self.expectThat(
            islice(pairwise(count()), 3),
            AfterPreprocessing(list, Equals([(0, 1), (1, 2), (2, 3)])),
        )



class ChainTests(TestCase):
    """
    Tests for ``Chain``.
    """
    def test_empty(self):
        """
        A ``Chain`` must have certificates.
        """
        self.assertRaises(
            InvariantException,
            lambda: Chain(certificates=Certificates([])),
        )


    def test_ordering(self):
        """
        Each certificate in ``Chain`` must be signed by the following certificate.
        """
        a_key, b_key, c_key = tuple(
            rsa.generate_private_key(
                public_exponent=65537,
                key_size=512,
                backend=default_backend(),
            )
            for i in range(3)
        )

        a_cert = cert(u"a.invalid", u"a.invalid", a_key.public_key(), a_key, True)
        b_cert = cert(u"a.invalid", u"b.invalid", b_key.public_key(), a_key, True)
        c_cert = cert(u"b.invalid", u"c.invalid", c_key.public_key(), b_key, False)

        a, b, c = pem.parse(b"\n".join(
            cert.public_bytes(serialization.Encoding.PEM)
            for cert
            in (a_cert, b_cert, c_cert)
        ))

        # a is not signed by b.  Rather, the reverse.  Therefore this ordering
        # is an error.
        self.expectThat(
            lambda: Chain(certificates=Certificates([c, a, b])),
            raises(InvariantException),
        )
        # c is signed by b and b is signed by a.  Therefore this is perfect.
        self.expectThat(
            Chain(certificates=Certificates([c, b, a])),
            IsInstance(Chain),
        )



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
A Kubernetes client.
"""

__all__ = [
    "version",
    "IObject", "IKubernetes", "IKubernetesClient",

    "KubernetesError", "UnrecognizedVersion", "UnrecognizedKind",

    "openapi_to_data_model", "v1_5_model",

    "memory_kubernetes",
    "network_kubernetes",  "network_kubernetes_from_context",
    "authenticate_with_serviceaccount",
    "authenticate_with_certificate",
    "authenticate_with_certificate_chain",

    # Deprecated.
    "v1", "v1beta1", "iobject_from_raw", "iobject_to_raw",
]

from incremental import Version

from ._metadata import version_tuple as _version_tuple
version = __version__ = Version("txkube", *_version_tuple)

from ._exception import KubernetesError, UnrecognizedVersion, UnrecognizedKind
from ._interface import IObject, IKubernetes, IKubernetesClient

from ._model import (
    openapi_to_data_model, v1_5_model,

    # Deprecated.
    iobject_from_raw,
    iobject_to_raw,
    v1, v1beta1,
)

from ._authentication import (
    authenticate_with_serviceaccount,
    authenticate_with_certificate,
    authenticate_with_certificate_chain,
)
from ._network import network_kubernetes, network_kubernetes_from_context
from ._memory import memory_kubernetes


def _deprecations():
    from twisted.python.deprecate import deprecatedModuleAttribute

    _0_2_0 = Version("txkube", 0, 2, 0)
    deprecatedModuleAttribute(
        _0_2_0,
        "See v1_5_model.v1 and IKubernetesClient.model.v1 instead.",
        "txkube",
        "v1",
    )
    deprecatedModuleAttribute(
        _0_2_0,
        "See v1_5_model.v1beta1 and IKubernetesClient.model.v1beta1 instead.",
        "txkube",
        "v1beta1",
    )
    deprecatedModuleAttribute(
        _0_2_0,
        "See v1_5_model.iobject_to_raw and "
        "IKubernetesClient.model.iobject_to_raw instead.",
        "txkube",
        "iobject_to_raw",
    )
    deprecatedModuleAttribute(
        _0_2_0,
        "See v1_5_model.iobject_from_raw and "
        "IKubernetesClient.model.iobject_from_raw instead.",
        "txkube",
        "iobject_from_raw",
    )

_deprecations()
del _deprecations



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Integration test generator for ``txkube.IKubernetesClient``.
"""

from re import search
from operator import attrgetter, setitem
from functools import partial, wraps
from itertools import repeat

from zope.interface.verify import verifyObject

from eliot import Message, start_action
from eliot.twisted import DeferredContext

from testtools.matchers import (
    AnyMatch, MatchesAll, MatchesStructure, Is, IsInstance, Equals, Not,
    Contains, AfterPreprocessing, MatchesPredicate, Always,
)

from twisted.python.compat import iteritems, unicode
from twisted.python.failure import Failure
from twisted.internet.defer import gatherResults
from twisted.internet.task import deferLater, cooperate
from twisted.web.http import NOT_FOUND, CONFLICT

from .._network import version_to_segments

from ..testing.matchers import MappingEquals, KubernetesObjectDeleted
from ..testing import TestCase, AsynchronousDeferredRunTest

from .. import (
    KubernetesError,
    IKubernetesClient,
)

from .strategies import (
    labels, creatable_namespaces, configmaps,
    deployments, replicasets, pods, services,
)



def async_test_runner():
    # This is a pretty long timeout but it includes sitting around waiting
    # for Kubernetes to finish deleting things so we get better test
    # isolation...  Namespace deletion, specifically, can be slow, with
    # some attempts taking 5 minutes or more.
    return AsynchronousDeferredRunTest.make_factory(timeout=600)



def matches_metadata(expected):
    return MatchesStructure(
        metadata=MatchesStructure(
            namespace=Equals(expected.namespace),
            name=Equals(expected.name),
            # TODO: It would be nice to compare labels but currently that
            # results in an annoying failure because of the confusion of
            # representation of the empty value - {} vs None.
            # https://github.com/LeastAuthority/txkube/issues/66
            # labels=Equals(expected.labels),
        ),
    )


def matches_namespace(ns):
    return matches_metadata(ns.metadata)


def matches_configmap(configmap):
    return matches_metadata(configmap.metadata)


def matches_deployment(deployment):
    return MatchesAll(
        matches_metadata(deployment.metadata),
        # Augh.  I think some kind of "expected is None or values match"
        # matcher would help?
        MatchesStructure(
            spec=MatchesStructure(
                selector=Equals(deployment.spec.selector),
                template=matches_metadata(deployment.spec.template.metadata),
            ),
        ),
    )


def matches_replicaset(replicaset):
    return MatchesAll(
        matches_metadata(replicaset.metadata),
    )


def matches_pod(pod):
    return MatchesAll(
        matches_metadata(pod.metadata),
    )


def matches_service(service):
    return matches_metadata(service.metadata)


def has_uid():
    return MatchesStructure(
        metadata=MatchesStructure(
            uid=Not(Equals(None)),
        ),
    )


def is_active():
    return MatchesStructure(
        status=MatchesStructure(
            phase=Equals(u"Active"),
        ),
    )


def different_name(name):
    if len(name) > 1:
        return name[:-1]
    return name + u"x"



def _named(model, kind, name, namespace=None):
    return kind(metadata=model.v1.ObjectMeta(name=name, namespace=namespace))



def poll(reactor, action, f, intervals):
    """
    Repeatedly call a function.

    :param IReactorTime reactor: The reactor to use for scheduling.

    :param action: An Eliot action in the context of which to call ``f``.

    :param f: A no-argument callable which returns a ``Deferred``.

    :param interfaces: An iterable of numbers giving the delay to use between
        each call to ``f``.  This also limits the total number of calls.  When
        the iterable is exhausted, ``poll`` will stop calling ``f``.

    :return: A ``Deferred`` that fires with any errors encountered by ``f`` or
        with ``StopIteration`` if the end of ``intervals`` is reached.
    """
    intervals = iter(intervals)
    def g():
        with action.context():
            Message.log(poll_iteration=None)
            # Not bothering with DeferredContext here because when
            # deferLater calls g, it can restore the action context from
            # the closure (as it just did a couple lines above).
            d = f()
            d.addCallback(
                lambda ignored: deferLater(reactor, next(intervals), g),
            )
            return d
    d = g()
    d.addCallback(lambda ignored: g())
    return d


_TRY_AGAIN = (
    u"the object has been modified; "
    u"please apply your changes to the latest version and try again"
)

def _replace(client, obj, transformation, retries):
    """
    Try to replace the given object with one that differs by the given
    transformation.

    Try again if the server reports the object has changed.
    """
    d = client.replace(obj.transform(*transformation))
    def maybe_outdated(reason):
        reason.trap(KubernetesError)
        # Uugghh...  This seems to be the only way to identify the condition.
        if _TRY_AGAIN in reason.value.status.message:
            d = client.get(obj)
            d.addCallback(lambda latest: _replace(
                client, latest, transformation, retries - 1
            ))
            return d
        return reason
    if retries > 0:
        d.addErrback(maybe_outdated)
    return d



def does_not_exist(client, obj):
    """
    Wait for ``obj`` to not exist.

    :param IKubernetesClient client: The client to use to determine existence.

    :param IObject obj: The object the existence of which to check.

    :return: A ``Deferred`` that fires when the object does not exist.
    """
    action = start_action(
        action_type=u"poll:start",
        obj=client.model.iobject_to_raw(obj),
    )
    with action.context():
        from twisted.internet import reactor
        d = poll(reactor, action, lambda: client.get(obj), repeat(0.5, 100))
        d = DeferredContext(d)
        # Once we get a NOT FOUND error, we're done.
        def trap_not_found(reason):
            reason.trap(KubernetesError)
            if reason.value.code == NOT_FOUND:
                return None
            return reason
        d.addErrback(trap_not_found)
        def trap_stop_iteration(reason):
            reason.trap(StopIteration)
            Message.log(
                does_not_exist=u"timeout",
                obj=client.model.iobject_to_raw(obj),
            )
            return None
        d.addErrback(trap_stop_iteration)
        return d.addActionFinish()


def needs(**to_create):
    """
    Create a function decorator which will create certain Kubernetes objects
    before calling the decorated function and delete them after it completes.

    This requires the decorated functions accept a first argument with a
    ``client`` attribute bound to an ``IKubernetesClient``.

    :param to_create: Keyword arguments with values of Hypothesis strategies
        which build ``IObject`` providers.  After being created, these objects
        will be passed to the decorated function using the same keyword
        arguments.

    :return: A function decorator.
    """
    def build_objects():
        objs = {}
        names = set()
        for name, strategy in iteritems(to_create):
            while True:
                obj = strategy.example()
                if obj.metadata.name not in names:
                    names.add(obj.metadata.name)
                    break
            objs[name] = obj
        return objs
    return _needs_objects(build_objects)



def _needs_objects(get_objects):
    def no_grace(model):
        return model.v1.DeleteOptions(gracePeriodSeconds=0)

    def decorator(f):
        @wraps(f)
        def wrapper(self, *a, **kw):
            # Get the objects we're supposed to create.
            to_create = get_objects()

            # Check to make sure there aren't keyword argument conflicts.  I
            # doubt this is an exhaustive safety check.
            overlap = set(to_create) & set(kw)
            if overlap:
                raise TypeError(
                    "Conflict between @needs() and **kw: {}".format(overlap)
                )

            def cleanup(obj):
                # Delete an object, if it exists, and wait for it to stop
                # existing.
                d = self.client.delete(obj, no_grace(self.model))
                # Some Kubernetes objects take a while to be deleted.
                # Specifically, Namespaces go though a slow "Terminating"
                # phase where things they contain are cleaned up.  Delay
                # cleanup completion until objects are *really* gone.
                d.addBoth(lambda ignored: does_not_exist(self.client, obj))
                return d

            def create(obj):
                # Create some object and arrange for it to be deleted later.
                d = self.client.create(obj)
                def created(result):
                    self.addCleanup(lambda: cleanup(result))
                    return result
                d.addCallback(created)
                return d

            # Keep track of the objects we created - keyed on name.
            created = {}

            # Create an object, deleting an existing one that's in the way
            # first, if necessary.
            def ensure(obj):
                d = cleanup(obj)
                d.addCallback(lambda ignored: create(obj))
                return d

            # Create the objects.
            task = cooperate(
                ensure(obj).addCallback(partial(setitem, created, name))
                for (name, obj)
                in sorted(to_create.items())
            )
            d = task.whenDone()

            # Call the decorated function.
            d.addCallback(lambda ignored: kw.update(created))
            d.addCallback(lambda ignored: f(self, *a, **kw))
            return d
        return wrapper
    return decorator



class _VersionTestsMixin(object):
    def test_version(self):
        """
        The server version can be retrieved using ``IKubernetesClient.version``.
        """
        d = self.client.version()
        d.addCallback(
            lambda version: self.assertThat(
                version, MatchesStructure(
                    major=IsInstance(unicode),
                    minor=IsInstance(unicode),
                ),
            ),
        )
        return d



class _NamespaceTestsMixin(object):
    @needs(obj=creatable_namespaces())
    def test_namespace(self, obj):
        """
        ``Namespace`` objects can be created and retrieved using the ``create``
        and ``list`` methods of ``IKubernetesClient``.
        """
        d = self.client.list(self.model.v1.Namespace)
        def check_namespaces(namespaces):
            self.assertThat(namespaces, IsInstance(self.model.v1.NamespaceList))
            # There are some built-in namespaces that we'll ignore.  If we
            # find the one we created, that's sufficient.
            self.assertThat(
                namespaces.items,
                AnyMatch(MatchesAll(matches_namespace(obj), has_uid(), is_active())),
            )
        d.addCallback(check_namespaces)
        return d


    @needs(obj=creatable_namespaces())
    def test_namespace_retrieval(self, obj):
        """
        A specific ``Namespace`` object can be retrieved by name using
        ``IKubernetesClient.get``.
        """
        return self._global_object_retrieval_by_name_test(
            obj,
            self.model.v1.Namespace,
            matches_namespace,
        )


    def test_namespace_replacement(self):
        """
        A specific ``Namespace`` object can be replaced by name using
        ``IKubernetesClient.replace``.
        """
        return self._object_replacement_test(
            None,
            creatable_namespaces(self.model),
            matches_namespace,
        )


    @needs(obj=creatable_namespaces())
    def test_namespace_deletion(self, obj):
        """
        ``IKubernetesClient.delete`` can be used to delete ``Namespace``
        objects.
        """
        d = self.client.delete(obj)
        def deleted_namespace(ignored):
            return self.client.list(self.model.v1.Namespace)
        d.addCallback(deleted_namespace)
        def check_namespaces(collection):
            active = list(
                ns.metadata.name
                for ns
                in collection.items
                if ns.status.phase == u"Active"
            )
            self.assertThat(
                active,
                Not(Contains(obj.metadata.name)),
            )
        d.addCallback(check_namespaces)
        return d


    def test_duplicate_namespace_rejected(self):
        """
        ``IKubernetesClient.create`` returns a ``Deferred`` that fails with
        ``KubernetesError`` if it is called with a ``Namespace`` object
        with the same name as a *Namespace* which already exists.
        """
        return self._create_duplicate_rejected_test(
            None,
            strategy=creatable_namespaces(self.model),
        )



class _ConfigMapTestsMixin(object):
    @needs(namespace=creatable_namespaces())
    def test_duplicate_configmap_rejected(self, namespace):
        """
        ``IKubernetesClient.create`` returns a ``Deferred`` that fails with
        ``KubernetesError`` if it is called with a ``ConfigMap`` object
        with the same name as a *ConfigMap* which already exists in the
        same namespace.
        """
        return self._create_duplicate_rejected_test(
            namespace,
            strategy=configmaps(self.model),
        )


    @needs(namespace=creatable_namespaces())
    def test_configmap(self, namespace):
        """
        ``ConfigMap`` objects can be created and retrieved using the ``create``
        and ``list`` methods of ``IKubernetesClient``.
        """
        self._create_list_test(
            namespace, configmaps(self.model), self.model.v1.ConfigMap,
            self.model.v1.ConfigMapList,
            matches_configmap,
        )


    def test_configmap_retrieval(self):
        """
        A specific ``ConfigMap`` object can be retrieved by name using
        ``IKubernetesClient.get``.
        """
        return self._namespaced_object_retrieval_by_name_test(
            configmaps(self.model),
            self.model.v1.ConfigMap,
            matches_configmap,
        )


    @needs(namespace=creatable_namespaces())
    def test_configmap_replacement(self, namespace):
        """
        A specific ``ConfigMap`` object can be replaced by name using
        ``IKubernetesClient.replace``.
        """
        return self._object_replacement_test(
            namespace,
            configmaps(self.model),
            matches_configmap,
        )


    def test_configmap_deletion(self):
        """
        A specific ``ConfigMap`` object can be deleted by name using
        ``IKubernetesClient.delete``.
        """
        return self._namespaced_object_deletion_by_name_test(
            configmaps(self.model),
            self.model.v1.ConfigMap,
        )


    ns_strategy = creatable_namespaces()
    @needs(ns_a=ns_strategy, ns_b=ns_strategy)
    def test_configmaps_sorted(self, ns_a, ns_b):
        """
        ``ConfigMap`` objects retrieved with ``IKubernetesClient.list`` appear in
        sorted order, with (namespace, name) as the sort key.
        """
        strategy = configmaps(self.model)
        objs = [
            strategy.example().transform(
                [u"metadata", u"namespace"], ns_a.metadata.name,
            ),
            strategy.example().transform(
                [u"metadata", u"namespace"], ns_b.metadata.name,
            ),
        ]
        d = gatherResults(list(self.client.create(obj) for obj in objs))
        def created_configmaps(ignored):
            return self.client.list(self.model.v1.ConfigMap)
        d.addCallback(created_configmaps)
        def check_configmaps(collection):
            self.expectThat(collection, items_are_sorted())
        d.addCallback(check_configmaps)
        return d



class _ReplicaSetTestsMixin(object):
    @needs(namespace=creatable_namespaces())
    def test_replicaset(self, namespace):
        """
        ``ReplicaSet`` objects can be created and retrieved using the ``create``
        and ``list`` methods of ``IKubernetesClient``.
        """
        return self._create_list_test(
            namespace, replicasets(self.model), self.model.v1beta1.ReplicaSet,
            self.model.v1beta1.ReplicaSetList, matches_replicaset,
        )


    @needs(namespace=creatable_namespaces())
    def test_duplicate_replicaset_rejected(self, namespace):
        """
        ``IKubernetesClient.create`` returns a ``Deferred`` that fails with
        ``KubernetesError`` if it is called with a ``ReplicaSet`` object with
        the same name as a *ReplicaSet* which already exists in the same
        namespace.
        """
        return self._create_duplicate_rejected_test(
            namespace,
            strategy=replicasets(self.model),
        )


    def test_replicaset_retrieval(self):
        """
        A specific ``ReplicaSet`` object can be retrieved by name using
        ``IKubernetesClient.get``.
        """
        return self._namespaced_object_retrieval_by_name_test(
            replicasets(self.model),
            self.model.v1beta1.ReplicaSet,
            matches_replicaset,
        )


    @needs(namespace=creatable_namespaces())
    def test_replicaset_replacement(self, namespace):
        """
        A specific ``ReplicaSet`` object can be replaced by name using
        ``IKubernetesClient.replace``.
        """
        return self._object_replacement_test(
            namespace,
            replicasets(self.model),
            matches_replicaset,
        )


    def test_replicaset_deletion(self):
        """
        A specific ``ReplicaSet`` object can be deleted by name using
        ``IKubernetesClient.delete``.
        """
        return self._namespaced_object_deletion_by_name_test(
            replicasets(self.model),
            self.model.v1beta1.ReplicaSet,
        )



class _PodTestsMixin(object):
    @needs(namespace=creatable_namespaces())
    def test_pod(self, namespace):
        """
        ``Pod`` objects can be created and retrieved using the ``create`` and
        ``list`` methods of ``IKubernetesClient``.
        """
        return self._create_list_test(
            namespace, pods(self.model), self.model.v1.Pod,
            self.model.v1.PodList, matches_pod,
        )


    @needs(namespace=creatable_namespaces())
    def test_duplicate_pod_rejected(self, namespace):
        """
        ``IKubernetesClient.create`` returns a ``Deferred`` that fails with
        ``KubernetesError`` if it is called with a ``Pod`` object with the
        same name as a *Pod* which already exists in the same namespace.
        """
        return self._create_duplicate_rejected_test(
            namespace,
            strategy=pods(self.model),
        )


    def test_pod_retrieval(self):
        """
        A specific ``Pod`` object can be retrieved by name using
        ``IKubernetesClient.get``.
        """
        return self._namespaced_object_retrieval_by_name_test(
            pods(self.model),
            self.model.v1.Pod,
            matches_pod,
        )


    @needs(namespace=creatable_namespaces())
    def test_pod_replacement(self, namespace):
        """
        A specific ``Pod`` object can be replaced by name using
        ``IKubernetesClient.replace``.
        """
        return self._object_replacement_test(
            namespace,
            pods(self.model),
            matches_pod,
        )


    def test_pod_deletion(self):
        """
        A specific ``Pod`` object can be deleted by name using
        ``IKubernetesClient.delete``.
        """
        return self._namespaced_object_deletion_by_name_test(
            pods(self.model),
            self.model.v1.Pod,
        )



class _DeploymentTestsMixin(object):
    @needs(namespace=creatable_namespaces())
    def test_deployment(self, namespace):
        """
        ``Deployment`` objects can be created and retrieved using the ``create``
        and ``list`` methods of ``IKubernetesClient``.
        """
        return self._create_list_test(
            namespace, deployments(self.model), self.model.v1beta1.Deployment,
            self.model.v1beta1.DeploymentList, matches_deployment,
        )


    @needs(namespace=creatable_namespaces())
    def test_duplicate_deployment_rejected(self, namespace):
        """
        ``IKubernetesClient.create`` returns a ``Deferred`` that fails with
        ``KubernetesError`` if it is called with a ``Deployment`` object with
        the same name as a *Deployment* which already exists in the same
        namespace.
        """
        return self._create_duplicate_rejected_test(
            namespace,
            strategy=deployments(self.model),
        )


    def test_deployment_retrieval(self):
        """
        A specific ``Deployment`` object can be retrieved by name using
        ``IKubernetesClient.get``.
        """
        return self._namespaced_object_retrieval_by_name_test(
            deployments(self.model),
            self.model.v1beta1.Deployment,
            matches_deployment,
        )


    @needs(namespace=creatable_namespaces())
    def test_deployment_replacement(self, namespace):
        """
        A specific ``Deployment`` object can be replaced by name using
        ``IKubernetesClient.replace``.
        """
        return self._object_replacement_test(
            namespace,
            deployments(self.model),
            matches_deployment,
        )


    def test_deployment_deletion(self):
        """
        A specific ``Deployment`` object can be deleted by name using
        ``IKubernetesClient.delete``.
        """
        return self._namespaced_object_deletion_by_name_test(
            deployments(self.model),
            self.model.v1beta1.Deployment,
        )



class _ServiceTestsMixin(object):
    @needs(namespace=creatable_namespaces())
    def test_service(self, namespace):
        """
        ``Service`` objects can be created and retrieved using the ``create`` and
        ``list`` methods of ``IKubernetesClient``.
        """
        self._create_list_test(
            namespace, services(self.model), self.model.v1.Service,
            self.model.v1.ServiceList,
            matches_service,
        )


    @needs(namespace=creatable_namespaces())
    def test_duplicate_service_rejected(self, namespace):
        """
        ``IKubernetesClient.create`` returns a ``Deferred`` that fails with
        ``KubernetesError`` if it is called with a ``ConfigMap`` object
        with the same name as a *ConfigMap* which already exists in the
        same namespace.
        """
        return self._create_duplicate_rejected_test(
            namespace,
            strategy=configmaps(self.model),
        )


    def test_service_retrieval(self):
        """
        A specific ``Service`` object can be retrieved by name using
        ``IKubernetesClient.get``.
        """
        return self._namespaced_object_retrieval_by_name_test(
            services(self.model),
            self.model.v1.Service,
            matches_service,
        )


    @needs(namespace=creatable_namespaces())
    def test_service_replacement(self, namespace):
        """
        A specific ``Service`` object can be replaced by name using
        ``IKubernetesClient.replace``.
        """
        return self._object_replacement_test(
            namespace,
            services(self.model),
            matches_service,
        )


    def test_service_deletion(self):
        """
        A specific ``Service`` object can be deleted by name using
        ``IKubernetesClient.delete``.
        """
        return self._namespaced_object_deletion_by_name_test(
            services(self.model),
            self.model.v1.Service,
        )



def kubernetes_client_tests(get_kubernetes):
    class KubernetesClientIntegrationTests(
        _VersionTestsMixin,
        _NamespaceTestsMixin,
        _ConfigMapTestsMixin,
        _DeploymentTestsMixin,
        _ReplicaSetTestsMixin,
        _PodTestsMixin,
        _ServiceTestsMixin,
        TestCase
    ):
        run_tests_with = async_test_runner()

        def setUp(self):
            super(KubernetesClientIntegrationTests, self).setUp()
            self.kubernetes = get_kubernetes(self)
            d = self.kubernetes.versioned_client()
            def got_client(client):
                self.client = client
                self.model = self.client.model
                self.addCleanup(self._cleanup)
            d.addCallback(got_client)
            return d


        def _cleanup(self):
            pool = getattr(self.client.agent, "_pool", None)
            if pool is None:
                return None
            from twisted.internet import reactor
            return gatherResults([
                pool.closeCachedConnections(),
                # Semi-work-around for
                # https://twistedmatrix.com/trac/ticket/8998
                deferLater(reactor, 1.0, lambda: None),
            ])


        def test_interfaces(self):
            """
            The client provides ``txkube.IKubernetesClient``.
            """
            verifyObject(IKubernetesClient, self.client)


        def test_not_found(self):
            """
            ``IKubernetesClient.list`` returns a ``Deferred`` that fails with
            ``KubernetesError`` when the server responds with an HTTP NOT
            FOUND status.
            """
            # Invent a type that the server isn't going to recognize.  This
            # could happen if we're talking to a server that is missing some
            # extension we thought it had, for example.
            class Mythical(object):
                apiVersion = u"v6txkube2"
                kind = u"Mythical"
                metadata = self.model.v1.ObjectMeta()

            # The client won't know where to route the request without this.
            # There is a more general problem here that the model is missing
            # some unpredictable information about where the relevant APIs are
            # exposed in the URL hierarchy.
            version_to_segments[Mythical.apiVersion] = (
                u"apis", u"extensions", Mythical.apiVersion,
            )
            self.addCleanup(lambda: version_to_segments.pop(Mythical.apiVersion))

            d = self.client.list(Mythical)
            def failed(reason):
                self.assertThat(reason, IsInstance(Failure))
                reason.trap(KubernetesError)
                self.assertThat(
                    reason.value,
                    MatchesStructure(
                        code=Equals(NOT_FOUND),
                        status=Equals(self.model.v1.Status(
                            metadata={},
                            status=u"Failure",
                            message=u"the server could not find the requested resource",
                            reason=u"NotFound",
                            details=dict(),
                            code=NOT_FOUND,
                        )),
                    ),
                )
            d.addBoth(failed)
            return d


        def _global_object_retrieval_by_name_test(self, obj, kind, matches):
            """
            Verify that a particular kind of non-namespaced Kubernetes object (such as
            *Namespace* or *PersistentVolume*) can be retrieved by name by
            calling ``IKubernetesClient.get`` with the ``IObject``
            corresponding to that kind as long as the object has its *name*
            metadata populated.
            """
            d = self.client.get(_named(self.model, kind, name=obj.metadata.name))
            def got_object(retrieved):
                self.assertThat(retrieved, matches(obj))
            d.addCallback(got_object)
            return d


        def _object_replacement_test(self, namespace, strategy, matches):
            """
            Verify that an existing object can be replaced by a new one.

            :param namespace: If the object belongs in a namespace, the name
                of an existing namespace to create it in.  ``None`` if not.

            :param strategy: A Hypothesis strategy for building the object to
                create and the object to replace it with.

            :param matches: A one-argument caller which takes the expected
                object and returns a testtools matcher for it.  Since created
                objects have some unpredictable server-generated fields, this
                matcher can compare just the important, predictable parts of
                the object.

            :return: A ``Deferred`` that fires when the behavior has been
                verified.
            """
            original = strategy.example()
            if namespace is not None:
                original = original.transform(
                    [u"metadata", u"namespace"], namespace.metadata.name,
                )
            # Different kinds of objects have different constraints on what
            # they can be replaced with.  To avoid running into those, make
            # the replacement just like the original but with some different
            # labels.  Hopefully this will always be allowed.
            #
            # Also, make sure _some_ labels are set to avoid triggering extra
            # server processing in some cases.  For example, Deployment
            # inherits labels from its template spec if its own labels are
            # empty.
            replacement_labels = labels().filter(lambda labels: len(labels)).example()

            d = self.client.create(original)
            def created(created):
                # Use the created object as the basis for the replacement
                # object.  Various server-populated fields (varying between
                # different kinds of objects) must match in the replacement.
                # Using our original object as the starting point will mess
                # these.
                transformation = (
                    [u"metadata", u"labels"],
                    replacement_labels,
                )
                return _replace(self.client, created, transformation, retries=2)
            d.addCallback(created)
            def replaced(result):
                self.expectThat(
                    result.metadata.labels,
                    MappingEquals(replacement_labels),
                )
                return self.client.get(original)
            d.addCallback(replaced)
            def retrieved(result):
                self.expectThat(
                    result.metadata.labels,
                    MappingEquals(replacement_labels),
                )
            d.addCallback(retrieved)
            return d


        def _create_list_test(self, namespace, strategy, cls, list_cls, matches):
            """
            Verify that a particular kind of namespaced Kubernetes object (such as
            *Service* or *Secret*) can be created using
            ``IKubernetesClient.create`` and then appears in the result of
            ``IKubernetesClient.list``.
            """
            # Get an object in the namespace.
            expected = strategy.example().transform(
                [u"metadata", u"namespace"],
                namespace.metadata.name,
            )
            d = self.client.create(expected)
            def created(actual):
                self.assertThat(actual, matches(expected))
                return self.client.list(cls)
            d.addCallback(created)
            def check(collection):
                self.assertThat(collection, IsInstance(list_cls))
                self.assertThat(collection.items, AnyMatch(matches(expected)))
            d.addCallback(check)
            return d


        def _create_duplicate_rejected_test(self, namespace, strategy):
            """
            Verify an object cannot be created if its name (and maybe namespace) is
            already taken.

            :param IObject obj: Some object to create and try to create again.

            :param unicode kind: The name of the collection of the *kind* of
                ``obj``, lowercase.  For example, *configmaps*.
            """
            obj = strategy.example()
            group = _infer_api_group(obj.apiVersion)

            # Put it in the namespace.
            if namespace is not None:
                obj = obj.transform(
                    [u"metadata", u"namespace"],
                    namespace.metadata.name,
                )
            d = self.client.create(obj)
            def created(ignored):
                return self.client.create(obj)
            d.addCallback(created)
            def failed(reason):
                self.assertThat(reason, IsInstance(Failure))
                reason.trap(KubernetesError)
                self.assertThat(
                    reason.value,
                    MatchesStructure(
                        code=Equals(CONFLICT),
                        status=MatchesStructure(
                            metadata=Equals(self.model.v1.ListMeta()),
                            status=Equals(u"Failure"),
                            # Error text is wonky and changes from version to
                            # version.  It's not a reliable source of
                            # information so don't bother making any
                            # particular assertions about its contents.
                            message=Always(),
                            reason=Equals(u"AlreadyExists"),
                            details=MatchesStructure.byEquality(
                                name=obj.metadata.name,
                                kind=obj.kind.lower() + u"s",
                                group=group,
                            ),
                            code=Equals(CONFLICT),
                        ),
                    ),
                )
            d.addBoth(failed)
            return d


        ns_strategy = creatable_namespaces()
        @needs(
            victim_namespace=ns_strategy,
            bystander_namespace=ns_strategy,
        )
        def _namespaced_object_deletion_by_name_test(
            self, strategy, cls, victim_namespace, bystander_namespace
        ):
            """
            Verify that a particular kind of namespaced Kubernetes object (such as
            *Deployment* or *Service*) can be deleted by name by calling
            ``IKubernetesClient.delete`` with the ``IObject`` corresponding to
            that kind as long as the object has its *name* and *namespace*
            metadata populated.

            :param strategy: A Hypothesis strategy for building the namespaced
                object to create and then retrieve.

            :param cls: The ``IObject`` implementation corresponding to the
                objects ``strategy`` can build.

            :param v1.Namespace victim_namespace: An existing namespace into
                which the probe object can be created (not part of the
                interface; value supplied by the decorator).

            :param v1.Namespace bystander_namespace: An existing namespace
                into which another object can be created (not part of the
                interface; value supplied by the decorator).

            :return: A ``Deferred`` that fires when the behavior has been
                verified.
            """
            # This object will be the target of the deletion.
            victim = strategy.example()
            # These bystanders should be left alone.
            bystander_a = strategy.example()
            bystander_b = strategy.example()
            victim = victim.transform(
                # Put the victim in the victim namespace.
                [u"metadata", u"namespace"], victim_namespace.metadata.name,
            )
            bystander_a = bystander_a.transform(
                # Put a bystander in the victim namespace, too.
                [u"metadata", u"namespace"], victim_namespace.metadata.name,
                # Also, be completely certain the bystander has a different
                # name.
                [u"metadata", u"name"], different_name(victim.metadata.name),
            )
            bystander_b = bystander_b.transform(
                # Put a bystander in another namespace too.
                [u"metadata", u"namespace"], bystander_namespace.metadata.name,
            )
            d = gatherResults(list(
                self.client.create(o)
                for o
                in [victim, bystander_a, bystander_b]
            ))
            def created_object(ignored):
                return self.client.delete(
                    _named(
                        self.model,
                        cls,
                        namespace=victim.metadata.namespace, name=victim.metadata.name,
                    ),
                    self.model.v1.DeleteOptions(gracePeriodSeconds=0),
                )
            d.addCallback(created_object)
            def deleted_object(result):
                self.expectThat(result, Is(None))
                return self.client.list(cls)
            d.addCallback(deleted_object)
            def listed_objects(collection):
                self.expectThat(
                    collection,
                    MatchesAll(
                        Not(KubernetesObjectDeleted(bystander_a)),
                        Not(KubernetesObjectDeleted(bystander_b)),
                        KubernetesObjectDeleted(victim),
                    ),
                )
            d.addCallback(listed_objects)
            return d


        @needs(namespace=creatable_namespaces())
        def _namespaced_object_retrieval_by_name_test(self, strategy, cls, matches, namespace):
            """
            Verify that a particular kind of namespaced Kubernetes object (such as
            *ConfigMap* or *PersistentVolumeClaim*) can be retrieved by name
            by calling ``IKubernetesClient.get`` with the ``IObject``
            corresponding to that kind as long as the object has its *name*
            and *namespace* metadata populated.

            :param strategy: A Hypothesis strategy for building the namespaced
                object to create and then retrieve.

            :param cls: The ``IObject`` implementation corresponding to the
                objects ``strategy`` can build.

            :param matches: A one-argument caller which takes the expected
                object and returns a testtools matcher for it.  Since created
                objects have some unpredictable server-generated fields, this
                matcher can compare just the important, predictable parts of
                the object.

            :param v1.Namespace namespace: An existing namespace into which
                the probe object can be created.

            :return: A ``Deferred`` that fires when the behavior has been
                verified.
            """
            obj = strategy.example()
            group = _infer_api_group(obj.apiVersion)
            # Move it to the namespace for this test.
            obj = obj.transform([u"metadata", u"namespace"], namespace.metadata.name)
            d = self.client.create(obj)
            def created_object(created):
                return self.client.get(_named(
                    self.model,
                    cls,
                    namespace=obj.metadata.namespace, name=obj.metadata.name,
                ))
            d.addCallback(created_object)
            def got_object(retrieved):
                self.expectThat(retrieved, matches(obj))
                # Try retrieving an object with the same name but a different
                # namespace.  We shouldn't find it.
                #
                # First, compute a legal but non-existing namespace name.
                bogus_namespace = different_name(obj.metadata.namespace)
                return self.client.get(
                    _named(
                        self.model,
                        cls,
                        namespace=bogus_namespace,
                        name=obj.metadata.name,
                    ),
                )
            d.addCallback(got_object)
            def check_error(reason):
                self.assertThat(reason, IsInstance(Failure))
                reason.trap(KubernetesError)
                self.assertThat(
                    reason.value,
                    MatchesStructure(
                        code=Equals(NOT_FOUND),
                        status=MatchesStructure(
                            metadata=Equals(self.model.v1.ListMeta()),
                            status=Equals(u"Failure"),
                            # Error text is wonky and changes from version to
                            # version.  It's not a reliable source of
                            # information so don't bother making any
                            # particular assertions about its contents.
                            message=Always(),
                            reason=Equals(u"NotFound"),
                            details=MatchesStructure.byEquality(
                                kind=obj.kind.lower() + u"s",
                                name=obj.metadata.name,
                                group=group,
                            ),
                            code=Equals(NOT_FOUND),
                        ),
                    ),
                )
            d.addBoth(check_error)
            return d

    return KubernetesClientIntegrationTests



def items_are_sorted():
    """
    Match an ObjectCollection if its items can be iterated in the Kubernetes
    canonical sort order - lexicographical by namespace, name.
    """
    def key(obj):
        return (
            getattr(obj.metadata, "namespace", None),
            obj.metadata.name,
        )

    def is_sorted(items, key):
        return list(items) == sorted(items, key=key)

    return AfterPreprocessing(
        attrgetter("items"),
        MatchesPredicate(
            partial(is_sorted, key=key),
            u"%s is not sorted by namespace, name",
        ),
    )



_CORE_PATTERNS = [
    # 1.6, 1.7
    u"^io.k8s.kubernetes.pkg.api.",
    u"^io.k8s.apimachinery.pkg.apis.meta.",

    # 1.5
    u"^v1$",
]



def _infer_api_group(apiVersion):
    for pattern in _CORE_PATTERNS:
        if search(pattern, apiVersion) is not None:
            return None
    return u"extensions"



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Tests for ``txkube._compat``.

"""

from .._compat import (
    native_string_to_bytes,
    native_string_to_unicode,
)

from ..testing import TestCase
from ..testing.matchers import raises_exception
from testtools.matchers import Equals



class NativeStringToBytesTests(TestCase):
    """
    Tests for ``native_string_to_bytes``.
    """
    def test_list(self):
        """
        Passing ``list`` as input should raise a ``TypeError``.
        """
        l = [1, 2, 3]
        self.assertThat(
            lambda: native_string_to_bytes(l),
            raises_exception(
                TypeError,
            ),
        )


    def test_none(self):
        """
        Passing ``None`` as input should raise a ``TypeError``.
        """
        self.assertRaises(
            TypeError,
            lambda: native_string_to_bytes(None),
        )


    def test_bytes_py2(self):
        """
        Passing ``bytes`` as input should return ``bytes``
        on Python 2.
        """
        if str is not bytes:
            # Python 3
            self.skipTest("skipping test on Python 3")

        self.assertThat(
            native_string_to_bytes(b"hello world"),
            Equals(b"hello world"),
        )


    def test_bytes_py3(self):
        """
        Passing ``bytes`` as input should return ``bytes`` should raise
        ``TypeError`` on Python 3.
        """
        if str is bytes:
            # Python 2
            self.skipTest("skipping test on Python 2")

        self.assertThat(
            lambda: native_string_to_bytes(b"hello world"),
            raises_exception(
                TypeError,
            ),
        )


    def test_unicode_py2(self):
        """
        Passing ``unicode`` as input should raise ``TypeError`` on Python 2.
        """
        if str is not bytes:
            # Python 3
            self.skipTest("skipping test on Python 3.")

        self.assertThat(
            lambda: native_string_to_bytes(u"hello world"),
            raises_exception(
                TypeError,
            ),
        )


    def test_unicode_py3(self):
        """
        Passing ``unicode`` as input should return ``unicode`` on Python 3.
        """
        if str is bytes:
            # Python 2
            self.skipTest(
                "native_string_to_bytes() does not accept unicode input on "
                "Python 2",
            )

        self.assertThat(
            native_string_to_bytes(u"hello world"),
            Equals(b"hello world"),
        )





class NativeStringToUnicodeTests(TestCase):
    """
    Tests for ``native_string_to_unicode``.
    """
    def test_list(self):
        """
        Passing ``list`` as input should raise a ``TypeError``.
        """
        l = [1, 2, 3]
        self.assertThat(
            lambda: native_string_to_unicode(l),
            raises_exception(
                TypeError,
            ),
        )


    def test_none(self):
        """
        Passing ``None`` as input should raise a ``TypeError``.
        """
        self.assertThat(
            lambda: native_string_to_unicode(None),
            raises_exception(
                TypeError,
            ),
        )


    def test_bytes_py2(self):
        """
        Passing ``bytes`` as input should return ``unicode`` on Python 2.
        """
        if str is not bytes:
            # Python 3
            self.skipTest("skipping test on Python 3")

        self.assertThat(
            native_string_to_unicode(b"hello world"),
            Equals(u"hello world"),
        )


    def test_bytes_py3(self):
        """
        Passing ``bytes`` as input should raise ``TypeError`` on Python 3.
        """
        if str is bytes:
            # Python 2
            self.skipTest("skipping test on Python 2")

        self.assertThat(
            lambda: native_string_to_unicode(b"hello world"),
            raises_exception(
                TypeError,
            ),
        )


    def test_unicode_py2(self):
        """
        Passing ``unicode`` as input should raise ``TypeError`` on Python 2.
        """
        if str is not bytes:
            # Python 3
            self.skipTest("skipping test on Python 3.")

        self.assertThat(
            lambda: native_string_to_unicode(u"hello world"),
            raises_exception(
                TypeError,
            ),
        )


    def test_unicode_py3(self):
        """
        Passing ``unicode`` as input should return ``unicode`` on Python 2.
        """
        if str is bytes:
            # Python 2
            self.skipTest("skipping test on Python 2")

        self.assertThat(
            native_string_to_unicode(u"hello world"),
            Equals(u"hello world"),
        )





# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Tests for ``txkube.testing._eliot``.
"""

from eliot import start_action, add_destinations, remove_destination

from testtools.matchers import (
    Contains,
    Equals,
)

from .. import TestCase

from .._eliot import (
    CaptureEliotLogs,
    _eliottree,
)


class EliotTreeTests(TestCase):
    """
    Tests for ``txkube.testing._eliot._eliottree``.
    """
    def test_tree(self):
        """
        ``_eliottree`` returns a ``bytes`` string containing a rendered tree of
        Eliot actions and messages.
        """
        events = []
        add_destinations(events.append)
        self.addCleanup(lambda: remove_destination(events.append))

        with start_action(action_type=u"foo"):
            pass

        # I don't know exactly what the tree rendering looks like.  That's why
        # I'm using eliot-tree!  So this assertion is sort of lame.
        self.assertThat(
            _eliottree(events),
            Contains(u"foo/1 \N{RIGHTWARDS DOUBLE ARROW} started"),
        )



class CaptureEliotLogsTests(TestCase):
    """
    Tests for ``txkube.testing._eliot.CaptureEliotLogs``.
    """
    def test_logs_as_detail(self):
        """
        Captured logs are available as details on the fixture.
        """
        fixture = CaptureEliotLogs()
        fixture.setUp()
        try:
            with start_action(action_type=u"foo"):
                pass
            details = fixture.getDetails()
        finally:
            fixture.cleanUp()
        self.assertThat(
            details[fixture.LOG_DETAIL_NAME].as_text(),
            Equals(_eliottree(fixture.logs)),
        )



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Tests for ``txkube.network_kubernetes``.

See ``get_kubernetes`` for pre-requisites.
"""

from os import (
    environ,
    pathsep,
)
from os.path import (
    expanduser,
)
from inspect import (
    getcallargs,
)
from base64 import b64encode

from zope.interface import implementer
from zope.interface.verify import verifyClass

import attr

from yaml import safe_dump

from fixtures import TempDir

from testtools.matchers import AnyMatch, ContainsDict, Equals
from testtools.twistedsupport import succeeded

from eliot.testing import capture_logging

from OpenSSL.crypto import FILETYPE_PEM

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend

from twisted.test.proto_helpers import MemoryReactorClock
from twisted.trial.unittest import TestCase as TwistedTestCase

from twisted.python.filepath import FilePath
from twisted.python.url import URL
from twisted.python.components import proxyForInterface
from twisted.internet.defer import Deferred, succeed
from twisted.internet.ssl import (
    CertificateOptions, DN, KeyPair, trustRootFromCertificates,
)
from twisted.internet.interfaces import (
    IReactorSSL, IReactorTCP, IReactorTime, IReactorPluggableNameResolver,
)
from twisted.internet.endpoints import SSL4ServerEndpoint
from twisted.web.client import Agent
from twisted.web.server import Site
from twisted.web.resource import Resource
from twisted.web.static import Data

from ..testing import TestCase, cert
from ..testing.integration import kubernetes_client_tests

from .. import (
    IObject, v1, network_kubernetes, network_kubernetes_from_context,
)

from .._network import (
    _Memo,
    collection_location,
    _merge_configs_from_env,
)

from .._compat import native_string_to_bytes


def get_kubernetes(case):
    """
    Create a real ``IKubernetes`` provider, taking necessary
    configuration details from the environment.

    To use this set ``TXKUBE_INTEGRATION_CONTEXT`` to a context in your
    ``kubectl`` configuration.  Corresponding details about connecting to a
    cluster will be loaded from that configuration.
    """
    try:
        context = environ["TXKUBE_INTEGRATION_CONTEXT"]
    except KeyError:
        case.skipTest("Cannot find TXKUBE_INTEGRATION_CONTEXT in environment.")
    else:
        from twisted.internet import reactor
        return network_kubernetes_from_context(reactor, context)



class MergeConfigsTests(TestCase):
    """
    Tests for ``_merge_configs_from_env``.
    """
    def setUp(self):
        super(TestCase, self).setUp()
        self.configs = FilePath(self.useFixture(TempDir()).path)


    def test_first_value_wins(self):
        """
        The first configuration file to set a particular value "wins".  The
        resulting merged configuration object has that value regardless of any
        values which might be set by later configuration files.
        """
        a = self.configs.child(b"a")
        b = self.configs.child(b"b")
        yaml = safe_dump({
            u"apiVersion": u"v1",
            u"kind": u"Config",
            u"current-context": u"foo",
        })
        a.setContent(native_string_to_bytes(yaml))
        yaml = safe_dump({
            u"apiVersion": u"v1",
            u"kind": u"Config",
            u"current-context": u"bar",
        })
        b.setContent(native_string_to_bytes(yaml))
        config = _merge_configs_from_env(u"{}:{}".format(
            a.asTextMode().path,
            b.asTextMode().path,
        ))
        self.assertEqual(
            u"foo",
            config.current_context,
        )


    def test_collections_merged(self):
        """
        Non-conflicting items in collections (clusters, users, contexts) are
        merged.
        """
        foo_context = {
            u"name": u"foo",
            u"context": {u"cluster": u"foo", u"user": u"foo"},
        }
        bar_context = {
            u"name": u"foo",
            u"context": {u"cluster": u"foo", u"user": u"foo"},
        }
        foo_user = {
            u"name": u"foo",
            u"user": {u"x-magic": u"foo"},
        }
        bar_user = {
            u"name": u"bar",
            u"user": {u"x-magic": u"bar"},
        }
        foo_cluster = {
            u"name": u"foo",
            u"cluster": {u"server": u"foo"},
        }
        bar_cluster = {
            u"name": u"bar",
            u"cluster": {u"server": u"bar"},
        }
        a = self.configs.child(b"a")
        b = self.configs.child(b"b")
        yaml = safe_dump({
            u"apiVersion": u"v1",
            u"kind": u"Config",
            u"contexts": [foo_context],
            u"users": [foo_user],
            u"clusters": [foo_cluster],
        })
        a.setContent(native_string_to_bytes(yaml))
        yaml = safe_dump({
            u"apiVersion": u"v1",
            u"kind": u"Config",
            u"contexts": [bar_context],
            u"users": [bar_user],
            u"clusters": [bar_cluster],
        })
        b.setContent(native_string_to_bytes(yaml))
        config = _merge_configs_from_env(u"{}:{}".format(
            a.asTextMode().path,
            b.asTextMode().path,
        ))
        self.expectThat(
            config.doc[u"contexts"],
            Equals([foo_context, bar_context]),
        )
        self.expectThat(
            config.doc[u"users"],
            Equals([foo_user, bar_user]),
        )
        self.expectThat(
            config.doc[u"clusters"],
            Equals([foo_cluster, bar_cluster]),
        )



class KubernetesClientIntegrationTests(kubernetes_client_tests(get_kubernetes)):
    """
    Integration tests which interact with a network-accessible
    Kubernetes deployment via ``txkube.network_kubernetes``.
    """



class CollectionLocationTests(TestCase):
    """
    Tests for ``collection_location``.
    """
    def _test_collection_location(self, version, kind, expected, namespace, instance):
        """
        Verify that ``collection_location`` for a particular version, kind,
        namespace, and Python object.

        :param unicode version: The *apiVersion* of the object to test.
        :param unicode kind: The *kind* of the object to test.

        :param tuple[unicode] expected: A representation of the path of the
            URL which should be produced.

        :param namespace: The namespace the Python object is to claim - a
            ``unicode`` string or ``None``.

        :param bool instance: Whether to make the Python object an instance
            (``True``) or a class (``False``)..
        """
        k = kind
        n = namespace
        @implementer(IObject)
        class Mythical(object):
            apiVersion = version
            kind = k

            metadata = v1.ObjectMeta(namespace=n)

            def serialize(self):
                return {}

        verifyClass(IObject, Mythical)

        if instance:
            o = Mythical()
        else:
            o = Mythical

        self.assertThat(
            collection_location(o),
            Equals(expected),
        )


    def test_v1_type(self):
        """
        ``collection_location`` returns a tuple representing an URL path like
        */api/v1/<kind>s* when called with an ``IObject`` implementation of a
        *v1* Kubernetes object kind.
        """
        self._test_collection_location(
            u"v1", u"Mythical", (u"api", u"v1", u"mythicals"),
            namespace=None,
            instance=False,
        )


    def test_v1_instance(self):
        """
        ``collection_location`` returns a tuple representing an URL path like
        */api/v1/namespace/<namespace>/<kind>s* when called with an
        ``IObject`` provider representing Kubernetes object of a *v1* kind.
        """
        self._test_collection_location(
            u"v1", u"Mythical",
            (u"api", u"v1", u"namespaces", u"ns", u"mythicals"),
            namespace=u"ns",
            instance=True,
        )


    def test_v1beta1_type(self):
        """
        ``collection_location`` returns a tuple representing an URL path like
        */apis/extensions/v1beta1/<kind>s* when called with an ``IObject``
        implementation of a *v1beta1* Kubernetes object kind.
        """
        self._test_collection_location(
            u"v1beta1", u"Mythical",
            (u"apis", u"extensions", u"v1beta1", u"mythicals"),
            namespace=None,
            instance=False,
        )


    def test_v1beta1_instance(self):
        """
        ``collection_location`` returns a tuple representing an URL path like
        */apis/extensions/v1beta1/<kind>s* when called with an ``IObject``
        implementation of a *v1beta1* Kubernetes object kind.
        """
        self._test_collection_location(
            u"v1beta1", u"Mythical",
            (u"apis", u"extensions", u"v1beta1", u"namespaces", u"ns",
             u"mythicals"),
            namespace=u"ns",
            instance=True,
        )


class ExtraNetworkClientTests(TestCase):
    """
    Direct tests for ``_NetworkClient`` that go beyond the guarantees of
    ``IKubernetesClient``.
    """
    @capture_logging(
        lambda self, logger: self.expectThat(
            logger.messages,
            AnyMatch(ContainsDict({
                u"action_type": Equals(u"network-client:list"),
                u"apiVersion": Equals(u"v1"),
                u"kind": Equals(u"Pod"),
            })),
        ),
    )
    def test_list_logging(self, logger):
        """
        ``_NetworkClient.list`` logs an Eliot event describing its given type.
        """
        client = network_kubernetes(
            base_url=URL.fromText(u"http://127.0.0.1/"),
            agent=Agent(MemoryReactorClock()),
        ).client()
        client.list(v1.Pod)



def self_signed_certificate_paths(key_path, cert_path, name):
    key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=1024,
        backend=default_backend(),
    )
    key_path.setContent(key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    ))

    certificate = cert(name, name, key.public_key(), key, True)
    cert_path.setContent(certificate.public_bytes(serialization.Encoding.PEM))

    return key_path, cert_path



class NetworkKubernetesFromContextTests(TwistedTestCase):
    """
    Direct tests for ``network_kubernetes_from_context``.
    """
    def test_client_chain_certificate(self):
        """
        A certificate chain in the *client-certificate* section of in the kube
        configuration file is used to configure the TLS context used when
        connecting to the API server.
        """
        def sign_ca_cert(key, requestObject, dn):
            from OpenSSL.crypto import X509, X509Extension
            from twisted.internet.ssl import Certificate

            req = requestObject.original
            cert = X509()
            dn._copyInto(cert.get_issuer())
            cert.set_subject(req.get_subject())
            cert.set_pubkey(req.get_pubkey())
            cert.gmtime_adj_notBefore(0)
            cert.gmtime_adj_notAfter(60 * 60)
            cert.set_serial_number(1)
            cert.add_extensions([
                X509Extension(b"basicConstraints", True, b"CA:TRUE"),
                # Not necessarily a good way to populate subjectAltName but it
                # quiets the deprecation warning we get from service_identity.
                X509Extension(b"subjectAltName", True, b"DNS:" + dn.commonName),
            ])
            cert.sign(key.original, "sha256")
            return Certificate(cert)

        ca_key = KeyPair.generate()
        ca_req = ca_key.requestObject(DN(commonName="kubernetes"))
        ca_cert = sign_ca_cert(ca_key, ca_req, DN(commonName="kubernetes"))

        intermediate_key = KeyPair.generate()
        intermediate_req = intermediate_key.requestObject(DN(commonName="intermediate"))
        intermediate_cert = sign_ca_cert(ca_key, intermediate_req, DN(commonName="kubernetes"))

        client_key = KeyPair.generate()
        client_req = client_key.requestObject(DN(commonName="client"))
        client_cert = intermediate_key.signRequestObject(DN(commonName="intermediate"), client_req, 1)

        chain = b"".join([
            client_cert.dumpPEM(),
            intermediate_cert.dumpPEM(),
        ])

        FilePath("ca.key").setContent(ca_key.dump(FILETYPE_PEM))
        FilePath("ca.crt").setContent(ca_cert.dump(FILETYPE_PEM))
        FilePath("intermediate.crt").setContent(intermediate_cert.dump(FILETYPE_PEM))
        FilePath("client.key").setContent(client_key.dump(FILETYPE_PEM))
        FilePath("client.crt").setContent(client_cert.dump(FILETYPE_PEM))
        FilePath("chain.crt").setContent(chain)

        config = self.write_config(ca_cert, chain, client_key)
        kubernetes = lambda reactor: network_kubernetes_from_context(
            reactor, "foo-ctx", path=config,
        )
        return self.check_tls_config(ca_key, ca_cert, kubernetes)


    def check_tls_config(self, ca_key, ca_cert, get_kubernetes):
        """
        Verify that a TLS server configured with the given key and certificate and
        the Kubernetes client returned by ``get_kubernetes`` can negotiate a
        TLS connection.
        """
        # Set up an HTTPS server that requires the certificate chain from the
        # configuration file.  This, because there's no way to pry inside a
        # Context and inspect its state nor any easy way to make Agent talk
        # over an in-memory transport.
        from twisted.internet import reactor
        endpoint = SSL4ServerEndpoint(
            reactor,
            0,
            CertificateOptions(
                privateKey=ca_key.original,
                certificate=ca_cert.original,
                trustRoot=trustRootFromCertificates([ca_cert]),
            ),
        )
        root = Resource()
        root.putChild(b"", Data(b"success", "text/plain"))

        # Construct the Kubernetes client objects with a Redirectable reactor.
        # This is necessary because the URL we pass to the Agent we get needs
        # to agree with the configuration file that was already written (or it
        # won't select the right client certificate).  Just one of the many
        # reasons it would be better if we didn't have to do real networking
        # here.
        redirectable = Redirectable(reactor)
        client = get_kubernetes(redirectable).client()
        agent = client.agent

        d = endpoint.listen(Site(root))
        def listening(port):
            self.addCleanup(port.stopListening)
            redirectable.set_redirect(port.getHost().host, port.getHost().port)
            url = b"https://127.0.0.1:8443/"
            return agent.request(b"GET", url)
        d.addCallback(listening)
        return d


    def write_config(self, ca_cert, chain, client_key):
        """
        Dump a kubectl config with the given details and return its location.

        :param Certificate ca_cert: The certificate authority certificate
            expected to have signed the server's certificate.

        :param unicode chain: PEM-encoded certificates starting with the
            client certificate and proceeding along a signature chain to a
            certificate signed by a certificate authority which the server
            recognizes.

        :param KeyPair client_key: The client's private key.

        :return FilePath: The path to the written configuration file.
        """
        config_file = FilePath(self.mktemp())
        config = safe_dump({
            "apiVersion": "v1",
            "kind": "Config",
            "contexts": [
                {
                    "name": "foo-ctx",
                    "context": {
                        "cluster": "foo-cluster",
                        "user": "foo-user",
                    },
                },
            ],
            "clusters": [
                {
                    "name": "foo-cluster",
                    "cluster": {
                        "certificate-authority-data": b64encode(ca_cert.dump(FILETYPE_PEM)),
                        "server": "https://127.0.0.1:8443/",
                    },
                },
            ],
            "users": [
                {
                    "name": "foo-user",
                    "user": {
                        "client-certificate-data": b64encode(chain),
                        "client-key-data": b64encode(client_key.dump(FILETYPE_PEM)),
                    },
                },
            ],
        })
        config_file.setContent(native_string_to_bytes(config))
        return config_file


    def test_kubeconfig_environment(self):
        """
        If ``KUBECONFIG`` is set in the environment,
        ``network_kubernetes_from_context`` merges the configuration from the
        files it refers to and returns an ``IKubernetes`` provider based on
        that configuration.
        """
        key_path, cert_path = self_signed_certificate_paths(
            FilePath(self.mktemp()),
            FilePath(self.mktemp()),
            u"x.invalid",
        )

        userauth = {
            "client-certificate": cert_path.path,
            "client-key": key_path.path,
        }

        a = FilePath(self.mktemp())
        yaml = safe_dump({
            "apiVersion": "v1",
            "kind": "Config",
            "contexts": [
                {"name": "a", "context": {"cluster": "a", "user": "a"}},
            ],
            "clusters": [{
                "name": "a",
                "cluster": {
                    "server": "https://a.example.com/",
                    "certificate-authority": cert_path.path,
                },
            }],
            "users": [
                {"name": "a", "user": userauth},
            ],
        })
        a.setContent(native_string_to_bytes(yaml))
        b = FilePath(self.mktemp())
        yaml = safe_dump({
            "apiVersion": "v1",
            "kind": "Config",
            "contexts": [
                {"name": "b", "context": {"cluster": "b", "user": "a"}},
            ],
            "clusters": [{
                "name": "b",
                "cluster": {
                    "server": "https://b.example.com/",
                    "certificate-authority": cert_path.path,
                },
            }],
            "users": [
                {"name": "b", "user": userauth},
            ],
        })
        b.setContent(native_string_to_bytes(yaml))
        c = FilePath(self.mktemp())
        yaml = safe_dump({
            "apiVersion": "v1",
            "kind": "Config",
            "current-context": "b",
        })
        c.setContent(native_string_to_bytes(yaml))
        kubeconfig = pathsep.join((a.path, b.path, c.path))
        kubernetes = network_kubernetes_from_context(
            MemoryReactorClock(),
            context=None,
            environ={"KUBECONFIG": kubeconfig},
        )
        self.assertEqual(
            "https://b.example.com/",
            kubernetes.base_url.asText(),
        )


    def test_default_os_environ(self):
        """
        If no value is passed for the ``environ`` parameter then
        ``network_kubernetes_from_context`` uses ``os.environ`` to look up the
        possible value of ``KUBECONFIG``.
        """
        key_path, cert_path = self_signed_certificate_paths(
            FilePath(self.mktemp()),
            FilePath(self.mktemp()),
            u"x.invalid",
        )
        userauth = {
            "client-certificate": cert_path.path,
            "client-key": key_path.path,
        }
        config = FilePath(self.mktemp())
        yaml = safe_dump({
            "apiVersion": "v1",
            "kind": "Config",
            "contexts": [
                {"name": "a", "context": {"cluster": "a", "user": "a"}},
            ],
            "clusters": [{
                "name": "a",
                "cluster": {
                    "server": "https://a.example.com/",
                    "certificate-authority": cert_path.path,
                },
            }],
            "users": [
                {"name": "a", "user": userauth},
            ],
        })
        config.setContent(native_string_to_bytes(yaml))
        import os
        self.patch(os, "environ", {u"KUBECONFIG": config.path})

        kubernetes = network_kubernetes_from_context(
            MemoryReactorClock(),
            context=u"a",
        )
        self.assertEqual(
            "https://a.example.com/",
            kubernetes.base_url.asText(),
        )


    def test_missing_kubeconfig(self):
        """
        When ``network_kubernetes_from_context`` is given no value for
        ``default_config_path`` it uses ``~/.kube/config`` as the value for
        that parameter.
        """
        callargs = getcallargs(network_kubernetes_from_context, MemoryReactorClock())
        self.assertEqual(
            FilePath(expanduser("~/.kube/config")),
            callargs["default_config_path"],
        )


    def test_default_config_path(self):
        """
        When ``network_kubernetes_from_context`` does not find ``KUBECONFIG`` in
        the environment it uses ``default_config_path`` as the path to the
        configuration file.
        """
        key_path, cert_path = self_signed_certificate_paths(
            FilePath(self.mktemp()),
            FilePath(self.mktemp()),
            u"x.invalid",
        )
        userauth = {
            "client-certificate": cert_path.path,
            "client-key": key_path.path,
        }
        config = FilePath(self.mktemp())
        yaml = safe_dump({
            "apiVersion": "v1",
            "kind": "Config",
            "contexts": [
                {"name": "a", "context": {"cluster": "a", "user": "a"}},
            ],
            "clusters": [{
                "name": "a",
                "cluster": {
                    "server": "https://a.example.com/",
                    "certificate-authority": cert_path.path,
                },
            }],
            "users": [
                {"name": "a", "user": userauth},
            ],
        })
        config.setContent(native_string_to_bytes(yaml))

        kubernetes = network_kubernetes_from_context(
            MemoryReactorClock(),
            context=u"a",
            environ={},
            default_config_path=config,
        )
        self.assertEqual(
            "https://a.example.com/",
            kubernetes.base_url.asText(),
        )



class MemoTests(TestCase):
    """
    Tests for ``_Memo``.
    """
    def test_get_empty(self):
        """
        In the empty state, ``get`` calls the given function and returns a
        ``Deferred`` that fires with that function's result.
        """
        result = object()
        def f():
            return succeed(result)

        m = _Memo()
        d = m.get(f)
        self.assertThat(d, succeeded(Equals(result)))


    def test_get_running(self):
        """
        In the running state, ``get`` does not call the given function and returns
        a ``Deferred`` that fires with the result of the function passed to
        the earlier ``get`` call.
        """
        result_obj = object()
        result = Deferred()
        def f1():
            return result
        f2 = None

        m = _Memo()
        d1 = m.get(f1)
        d2 = m.get(f2)
        result.callback(result_obj)
        self.assertThat(d1, succeeded(Equals(result_obj)))
        self.assertThat(d2, succeeded(Equals(result_obj)))


    def test_get_value(self):
        """
        In the value state, ``get`` does not call the given function and returns a
        ``Deferred`` that fires with the result of the function passed to the
        earlier ``get`` call.
        """
        result_obj = object()
        result = succeed(result_obj)
        def f1():
            return result
        f2 = None

        m = _Memo()
        d1 = m.get(f1)
        d2 = m.get(f2)
        self.assertThat(d1, succeeded(Equals(result_obj)))
        self.assertThat(d2, succeeded(Equals(result_obj)))



@attr.s
class Redirectable(proxyForInterface(IReactorSSL),
                   proxyForInterface(IReactorTCP),
                   proxyForInterface(IReactorTime),
                   proxyForInterface(IReactorPluggableNameResolver)):
    """
    A reactor which ignores the requested destination and always
    connects to an alternate address instead.

    :ivar host: The host portion of the alternate address.
    :ivar port: The host portion of the alternate address.
    """
    original = attr.ib()

    def set_redirect(self, host, port):
        """
        Specify the alternate address to which connections will be directed.
        """
        self.host, self.port = host, port


    def connectSSL(self, host, port, *a, **kw):
        """
        Establish a TLS connection to the alternate address instead of the given
        address.
        """
        return self.original.connectSSL(self.host, self.port, *a, **kw)


    def connectTCP(self, host, port, *a, **kw):
        """
        Establish a TCP connection to the alternate address instead of the given
        address.
        """
        return self.original.connectTCP(self.host, self.port, *a, **kw)



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
xUnit TestCase for txkube testing.
"""

from os import environ

from hypothesis import (
    HealthCheck,
    settings,
)

from fixtures import CompoundFixture

from testtools import TestCase as TesttoolsTestCase
from testtools.twistedsupport import AsynchronousDeferredRunTest

from twisted.python.failure import Failure

from ._eliot import CaptureEliotLogs


def _setup_hypothesis():
    settings.register_profile(
        "ci",
        suppress_health_check=[
            # CPU resources available to CI builds typically varies
            # significantly from run to run making it difficult to determine
            # if "too slow" data generation is a result of the code or the
            # execution environment.  Prevent these checks from
            # (intermittently) failing tests that are otherwise fine.
            HealthCheck.too_slow,
        ],
        # By the same reasoning as above, disable the deadline check.
        deadline=None,
    )
    settings.load_profile(environ.get("TXKUBE_HYPOTHESIS_PROFILE", "default"))
_setup_hypothesis()


class AsynchronousDeferredRunTest(AsynchronousDeferredRunTest):
    """
    An asynchronous runner supporting Eliot.
    """
    def _get_log_fixture(self):
        """
        Add ``CaptureEliotLogs`` to the log fixtures which receive special
        treatment so as to be "cleaned up" in the timeout case.

        This ensures eliot logs are reported when tests time out - something
        that will not happen using the normal ``useFixture`` API.

        See <https://bugs.launchpad.net/testtools/+bug/897196>.
        """
        return CompoundFixture([
            super(AsynchronousDeferredRunTest, self)._get_log_fixture(),
            CaptureEliotLogs(),
        ])



class TestCase(TesttoolsTestCase):
    """
    A base class for test cases which automatically uses the
    ``CaptureEliotLogs`` fixture.
    """
    # expectThat and Hypothesis don't communicate well about when the
    # test has failed.  Give them a little help.  These two Hypothesis
    # hooks will check for a flag that testtools sets when it thinks
    # the test has failed and turn it into something Hypothesis can
    # recognize.
    def setup_example(self):
        try:
            # TesttoolsTestCase starts without this attribute set at all.  Get
            # us back to that state.  It won't be set at all on the first
            # setup_example call, nor if the previous run didn't have a failed
            # expectation.
            del self.force_failure
        except AttributeError:
            pass

    def teardown_example(self, ignored):
        if getattr(self, "force_failure", False):
            # If testtools things it's time to stop, translate this into a
            # test failure exception that Hypothesis can see.  This lets
            # Hypothesis know when it has found a falsifying example.  Without
            # it, Hypothesis can't see which of its example runs caused
            # problems.
            self.fail("expectation failed")



def assertNoResult(case, d):
    """
    Assert that ``d`` does not have a result at this point.
    """
    result = []
    d.addBoth(result.append)
    if result:
        if isinstance(result[0], Failure):
            result[0].raiseException()
        else:
            case.fail("Got {} but expected no result".format(result[0]))



# Copyright Least Authority Enterprises.
# See LICENSE for details.

__all__ = [
    "strategies",
    "integration",
    "TestCase", "AsynchronousDeferredRunTest",
    "assertNoResult",
    "cert",
]

from ._testcase import (
    TestCase,
    AsynchronousDeferredRunTest,
    assertNoResult,
)

from cryptography.x509 import (
    CertificateBuilder,
    SubjectAlternativeName,
    BasicConstraints,
    DNSName,
    Name,
    NameAttribute,
    random_serial_number,
)
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.backends import default_backend

from datetime import datetime, timedelta

def cert(issuer, subject, pubkey, privkey, ca):
    builder = CertificateBuilder(
    ).issuer_name(
        Name([NameAttribute(NameOID.COMMON_NAME, issuer)]),
    ).subject_name(
        Name([NameAttribute(NameOID.COMMON_NAME, subject)]),
    ).add_extension(
        SubjectAlternativeName([DNSName(subject)]),
        critical=False,
    )
    if ca:
        builder = builder.add_extension(
            BasicConstraints(True, None),
            critical=True,
        )
    return builder.public_key(pubkey,
    ).serial_number(random_serial_number(),
    ).not_valid_before(datetime.utcnow(),
    ).not_valid_after(datetime.utcnow() + timedelta(seconds=1),
    ).sign(privkey, SHA256(), default_backend(),
    )



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Helpers for Python 2/3 compatibility.
"""

from twisted.python.compat import _PY3

def encode_environ(env):
    """
    Convert a ``dict`` of ``unicode`` keys and values to ``bytes`` on Python 2,
    but return the ``dict`` unmodified on Python 3.
    """
    if _PY3:
        return env
    else:
        bytes_env = {}
        for key in env:
            bytes_env[key.encode("ascii")] = env[key].encode("ascii")
        return bytes_env



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Tests for ``txkube._model``.
"""

from json import loads, dumps

from zope.interface import implementer
from zope.interface.verify import verifyObject

import attr

from pyrsistent import (
    InvariantException,
    freeze,
)

from testtools.matchers import (
    NotEquals,
    Equals,
    LessThan,
    GreaterThan,
    MatchesStructure,
    Not,
    Is,
    Contains,
    ContainsAll,
    raises,
    IsInstance,
)
from testtools.twistedsupport import succeeded

from hypothesis import HealthCheck, settings, given, assume
from hypothesis.strategies import sampled_from, choices

from twisted.python.failure import Failure
from twisted.internet.defer import gatherResults
from twisted.web.iweb import IResponse
from twisted.web.http_headers import Headers
from twisted.web.client import ResponseDone

from ..testing import TestCase
from ..testing.matchers import (
    PClassEquals,
    EqualElements,
    MappingEquals,
    raises_exception,
)
from ..testing.strategies import (
    iobjects,
    namespacelists,
    objectcollections,
)

from .. import (
    UnrecognizedVersion, UnrecognizedKind,
    KubernetesError,
    IObject,
    v1_5_model, openapi_to_data_model,
)

from .._model import set_if_none

from .._compat import dumps_bytes



def models():
    return sampled_from([v1_5_model])



class SerializationTests(TestCase):
    """
    Tests for ``iobject_to_raw`` and ``iobject_from_raw``.
    """
    def test_v1_apiVersion(self):
        """
        Objects from ``v1`` serialize with an *apiVersion* of ``u"v1"``.
        """
        model = v1_5_model

        obj = model.v1.ComponentStatus()
        raw = model.iobject_to_raw(obj)
        self.expectThat(
            raw[u"apiVersion"],
            Equals(u"v1"),
        )
        self.expectThat(
            model.iobject_from_raw(raw),
            IsInstance(model.v1.ComponentStatus),
        )


    def test_v1beta1_apiVersion(self):
        """
        Objects from ``v1beta1`` serialize with an *apiVersion* of
        ``u"extensions/v1beta1"``.
        """
        model = v1_5_model

        obj = model.v1beta1.CertificateSigningRequest()
        raw = model.iobject_to_raw(obj)
        self.expectThat(
            raw[u"apiVersion"],
            Equals(u"extensions/v1beta1"),
        )
        self.expectThat(
            model.iobject_from_raw(raw),
            IsInstance(model.v1beta1.CertificateSigningRequest),
        )



class IObjectTests(TestCase):
    """
    Tests for ``IObject``.
    """
    @given(obj=iobjects())
    def test_interface(self, obj):
        """
        The object provides ``IObject``.
        """
        verifyObject(IObject, obj)


    @given(models())
    def test_constant_attributes(self, model):
        """
        The ``apiVersion`` and ``kind`` attributes reflect the Kubernetes object
        apiVersion and kind fields.
        """
        p = model.v1.Pod()
        self.expectThat(p.apiVersion, Equals(u"v1"))
        self.expectThat(p.kind, Equals(u"Pod"))

        pl = model.v1.PodList()
        self.expectThat(pl.apiVersion, Equals(u"v1"))
        self.expectThat(pl.kind, Equals(u"PodList"))

        d = model.v1beta1.Deployment()
        self.expectThat(d.apiVersion, Equals(u"v1beta1"))
        self.expectThat(d.kind, Equals(u"Deployment"))

        dl = model.v1beta1.DeploymentList()
        self.expectThat(dl.apiVersion, Equals(u"v1beta1"))
        self.expectThat(dl.kind, Equals(u"DeploymentList"))


    @given(obj=iobjects())
    def test_serialization_roundtrip(self, obj):
        """
        An ``IObject`` provider can be round-trip through JSON using
        ``iobject_to_raw`` and ``iobject_from_raw``.
        """
        model = v1_5_model

        marshalled = model.iobject_to_raw(obj)

        # Every IObject has these marshalled fields - and when looking at the
        # marshalled form, they're necessary to figure out the
        # schema/definition for the data.  We can't say anything in general
        # about the *values* (because of things like "extensions/v1beta1") but
        # we can at least assert the keys are present.
        self.expectThat(marshalled, ContainsAll([u"kind", u"apiVersion"]))

        # We should be able to unmarshal the data back to the same model
        # object as we started with.
        reloaded = model.iobject_from_raw(marshalled)
        self.expectThat(obj, PClassEquals(reloaded))

        # And, to be extra sure (ruling out any weird Python object
        # semantic hijinx), that that reconstituted object should marshal
        # back to exactly the same simplified object graph.
        remarshalled = model.iobject_to_raw(reloaded)
        self.expectThat(marshalled, MappingEquals(remarshalled))

        # Also, the marshalled form must be JSON compatible.
        serialized = dumps(marshalled)
        deserialized = loads(serialized)
        self.expectThat(marshalled, MappingEquals(deserialized))


    @given(objectcollections())
    def test_empty_collection(self, collection):
        """
        The ``items`` of a collection can be made empty in a couple different
        ways.
        """
        self.expectThat(collection.set(items=None).items, Equals([]))
        self.expectThat(collection.set(items=[]).items, Equals([]))



    @given(collection=namespacelists(), choose=choices())
    def test_unique_contents(self, collection, choose):
        """
        A collection type cannot contain more than one object with a particular
        namespace / name pair.
        """
        assume(len(collection.items) > 0)
        item = choose(collection.items)
        self.expectThat(
            lambda: collection.add(item),
            raises_exception(InvariantException),
        )


    def test_unknown_version(self):
        """
        ``iobject_from_raw`` raises ``UnrecognizedVersion`` if it does not
        recognize the *apiVersion* in the given data.
        """
        model = v1_5_model

        obj = {
            u"apiVersion": u"invalid.example.txkube",
            u"kind": u"Service",
        }
        self.assertThat(
            lambda: model.iobject_from_raw(obj),
            raises(UnrecognizedVersion(obj[u"apiVersion"], obj)),
        )


    def test_unknown_kind(self):
        """
        ``iobject_from_raw`` raises ``UnrecognizedKind`` if it does not recognize
        the *kind* in the given data.
        """
        model = v1_5_model

        obj = {
            u"apiVersion": u"v1",
            u"kind": u"SomethingFictional",
        }
        self.assertThat(
            lambda: model.iobject_from_raw(obj),
            raises(UnrecognizedKind(u"v1", u"SomethingFictional", obj)),
        )



class NamespaceTests(TestCase):
    """
    Other tests for ``Namespace``.
    """
    @given(models())
    def test_default(self, model):
        """
        ``Namespace.default`` returns the *default* namespace.
        """
        self.assertThat(
            model.v1.Namespace.default(),
            MatchesStructure(
                metadata=MatchesStructure(
                    name=Equals(u"default"),
                ),
            ),
        )


    @given(models())
    def test_fill_defaults(self, model):
        """
        ``Namespace.fill_defaults`` returns a ``Namespace`` with *uid* metadata
        and an active *status*.
        """
        # If they are not set already, a uid is generated and put into the
        # metadata and the status is set to active.
        sparse = model.v1.Namespace(metadata=model.v1.ObjectMeta(name=u"foo"))
        filled = sparse.fill_defaults()
        self.expectThat(
            filled,
            MatchesStructure(
                metadata=MatchesStructure(
                    uid=Not(Is(None)),
                ),
                status=Equals(model.v1.NamespaceStatus.active()),
            ),
        )



class NamespaceListTests(TestCase):
    """
    Tests for ``NamespaceList``.
    """
    @settings(suppress_health_check=[HealthCheck.exception_in_generation])
    @given(collection=namespacelists(), choose=choices())
    def test_remove(self, collection, choose):
        """
        ``NamespaceList.remove`` creates a new ``NamespaceList`` which does not
        have the given item.
        """
        assume(len(collection.items) > 0)
        item = choose(collection.items)
        removed = collection.remove(item)
        self.assertThat(removed.items, Not(Contains(item)))


    @given(collection=namespacelists(), choose=choices())
    def test_item_by_name(self, collection, choose):
        """
        ``NamespaceList.item_by_name`` returns the ``Namespace`` with the matching
        name.
        """
        assume(len(collection.items) > 0)
        for item in collection.items:
            self.expectThat(collection.item_by_name(item.metadata.name), Is(item))

        item = choose(collection.items)
        collection = collection.remove(item)
        self.expectThat(
            lambda: collection.item_by_name(item.metadata.name),
            raises(KeyError(item.metadata.name)),
        )

    @given(collection=namespacelists(), choose=choices())
    def test_no_duplicates(self, collection, choose):
        assume(len(collection.items) > 0)
        self.expectThat(
            lambda: collection.add(choose(collection.items)),
            raises_exception(InvariantException),
        )


    @given(collection=namespacelists())
    def test_constant_attributes(self, collection):
        self.expectThat(collection.kind, Equals(u"NamespaceList"))
        self.expectThat(collection.apiVersion, Equals(u"v1"))



class SetIfNoneTests(TestCase):
    """
    Tests for ``set_if_none``.
    """
    def test_none(self):
        """
        If the value for transformation is ``None``, the result contains the new
        value instead.
        """
        structure = freeze({u"foo": None})
        transformed = structure.transform([u"foo"], set_if_none(u"bar"))
        self.assertThat(transformed[u"foo"], Equals(u"bar"))


    def test_not_none(self):
        """
        If the value for transformation is not ``None``, the result contains the
        original value.
        """
        structure = freeze({u"foo": u"baz"})
        transformed = structure.transform([u"foo"], set_if_none(u"bar"))
        self.assertThat(transformed[u"foo"], Equals(u"baz"))



@implementer(IResponse)
@attr.s
class MemoryResponse(object):
    version = attr.ib()
    code = attr.ib()
    phrase = attr.ib()
    headers = attr.ib()
    length = attr.ib()

    request = attr.ib()
    previousResponse = attr.ib()

    _body = attr.ib()

    def deliverBody(self, protocol):
        protocol.makeConnection(None)
        protocol.dataReceived(self._body)
        protocol.connectionLost(Failure(ResponseDone()))



class KubernetesErrorTests(TestCase):
    """
    Tests for ``KubernetesError``.
    """
    def test_from_response(self):
        """
        ``from_response`` returns the same value as ``from_model_and_response``
        when called with the v1.5 model.
        """
        def response():
            body = dumps_bytes(v1_5_model.iobject_to_raw(v1_5_model.v1.Status()))
            return MemoryResponse(
                version=(b"HTTP", 1, 1),
                code=200,
                phrase=b"OK",
                headers=Headers(),
                length=len(body),
                request=None,
                previousResponse=None,
                body=body,
            )

        ds = gatherResults([
            KubernetesError.from_response(response()),
            KubernetesError.from_model_and_response(v1_5_model, response()),
        ])
        self.assertThat(
            ds,
            succeeded(EqualElements()),
        )


    def test_comparison(self):
        """
        The binary comparison operations work on ``KubernetesError`` as expected.
        """
        model = v1_5_model

        a1 = KubernetesError(200, model.v1.Status(status=u"A"))
        a2 = KubernetesError(200, model.v1.Status(status=u"A"))
        b = KubernetesError(201, model.v1.Status(status=u"A"))
        c = KubernetesError(200, model.v1.Status(status=u"B"))

        # a1 == a2
        self.expectThat(a1, Equals(a2))
        # not (a1 != a2)
        self.expectThat(a1, Not(NotEquals(a2)))
        # not (a1 > a2)
        self.expectThat(a1, Not(GreaterThan(a2)))
        # not (a1 < a2)
        self.expectThat(a1, Not(LessThan(a2)))

        # not (a1 == b)
        self.expectThat(a1, Not(Equals(b)))
        # a1 != b
        self.expectThat(a1, NotEquals(b))
        # a1 < b
        self.expectThat(a1, LessThan(b))
        # not (a1 > b)
        self.expectThat(a1, Not(GreaterThan(b)))

        # not (a1 == c)
        self.expectThat(a1, Not(Equals(b)))
        # a1 != c
        self.expectThat(a1, NotEquals(b))
        # a1 < c
        self.expectThat(a1, LessThan(c))
        # not (a1 > c)
        self.expectThat(a1, Not(GreaterThan(c)))

        largest = KubernetesError(999, model.v1.Status(status=u"Z"))
        smallest = KubernetesError(1, model.v1.Status(status=u"A"))

        # a1 < largest
        self.expectThat(a1, LessThan(largest))
        # a1 > smallest
        self.expectThat(a1, GreaterThan(smallest))


    def test_equal_or_not_equal_to_other_type(self):
        """
        When comparing against something which is not a `KubernetesError`
        '==' returns `False` and '!=' returns `True`.
        """
        model = v1_5_model
        a1 = KubernetesError(200, model.v1.Status(status=u"A"))
        self.expectThat(
            a1 == 35,
            Is(False),
        )
        self.expectThat(
            a1 != 35,
            Is(True),
        )


    def test_greater_or_less_than_other_type_py3(self):
        """
        '<' and '>' binary comparisons raise `TypeError` on Python 3, when
        compared with something which is not a `KubernetesError`.
        """
        try:
            cmp
            self.skipTest("skipping test on Python 2")
        except NameError:
            # Python 3 does not have cmp()
            pass

        model = v1_5_model
        a1 = KubernetesError(200, model.v1.Status(status=u"A"))
        self.expectThat(
            lambda: a1 < 35,
            raises_exception(TypeError),
        )
        self.expectThat(
            lambda: a1 > 35,
            raises_exception(TypeError),
        )



class Extra15DataModelTests(TestCase):
    """
    Tests for handling of certain Kubernetes 1.5 Swagger specifications.
    """
    def test_status(self):
        """
        If the Kubernetes-reported Swagger specification is missing the Status
        definitions, ``openapi_to_data_model`` returns a data model that
        defines them anyway.
        """
        openapi = v1_5_model.spec.to_document()
        openapi[u"definitions"].clear()
        model = openapi_to_data_model(openapi)
        model.v1.Status
        model.v1.StatusCause
        model.v1.StatusDetails
        model.v1.ListMeta
